# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import json
import logging
import os
import shutil

from termcolor import colored

from climetlab.core.settings import SETTINGS
from climetlab.utils import humanize, tqdm
from climetlab.utils.dates import to_datetime

from .tools import parse_args, print_table

LOG = logging.getLogger(__name__)

EPILOG = """
SIZE can be expressed using suffixes such a K, M, G, etc. For example
``--larger 1G`` will match all cache entries larger than 1 GiB.

DATE can be expressed as absolute time like ``2021-10-10T22:59:00``` or relative such
as ``1h`` (one hour ago) or ``2d`` (two days ago).

The ``--older`` and ``--newer`` consider the *creation date* of
cache entries, unless ``--accessed`` is specified. In this case the time of
*last access* is used.

Example, to remove large files not accessed for one week:

   ``decache --accessed --older 1w --larger 1G``

"""

MATCHER = dict(
    epilog=EPILOG,
    match=dict(type=str, metavar="STRING", help="TODO"),
    newer=dict(type=str, metavar="DATE", help="TODO"),
    older=dict(type=str, metavar="DATE", help="TODO"),
    accessed=dict(
        action="store_true",
        help="use the date of last access instead of the creation date",
    ),
    larger=dict(
        type=str,
        metavar="SIZE",
        help="consider only cache entries that are larger than SIZE bytes",
    ),
    smaller=dict(
        type=str,
        metavar="SIZE",
        help="consider only cache entries that are smaller than SIZE bytes",
    ),
)


def parse_size(txt):
    return humanize.as_bytes(txt)


def parse_user_date(value):
    try:
        return to_datetime(value)
    except ValueError:
        return datetime.datetime.now() - humanize.as_timedelta(value)


class Matcher:
    def __init__(self, args):
        self.message = []

        for k in MATCHER.keys():
            if k != "epilog":
                if isinstance(args, dict):
                    value = args.get(k, None)
                else:
                    value = getattr(args, k)
                setattr(self, k, value)

        if self.match is not None:
            self.message.append(f"matching '{self.match}'")

        if self.newer is not None:
            self.newer = parse_user_date(self.newer)
            value = humanize.rounded_datetime(self.newer)
            self.message.append(f"newer than '{value}'")

        if self.older is not None:
            self.older = parse_user_date(self.older)
            value = humanize.rounded_datetime(self.older)
            self.message.append(f"older than '{value}'")

        if self.smaller is not None:
            self.smaller = parse_size(self.smaller)
            value = humanize.bytes(self.smaller)
            self.message.append(f"smaller than {value}")

        if self.larger is not None:
            self.larger = parse_size(self.larger)
            value = humanize.bytes(self.larger)
            self.message.append(f"larger than {value}")

        self.undefined = len(self.message) == 0

    def __call__(self, entry):
        if self.undefined:
            return True

        if self.smaller is not None:
            if entry["size"] is None or entry["size"] > self.smaller:
                return False

        if self.larger is not None:
            if entry["size"] is None or entry["size"] < self.larger:
                return False

        creation_date = to_datetime(entry["creation_date"])

        if self.newer is not None:
            if creation_date < self.newer:
                return False

        if self.older is not None:
            if creation_date > self.older:
                return False

        if self.match is not None:
            if not self._match(entry):
                return False

        # accesses
        # last_access

        return True

    def _match(self, entry):
        if isinstance(entry, (list, tuple)):
            return any(self._match(x) for x in entry)

        if isinstance(entry, dict):
            return any(self._match(x) for x in entry.values())

        return self.match in str(entry)


class CacheCmd:
    @parse_args(
        json=dict(action="store_true", help="produce a JSON output"),
        all=dict(action="store_true"),
        path=dict(
            action="store_true", help="print the path of cache directory and exit"
        ),
        sort=dict(
            type=str,
            metavar="KEY",
            help="sort output according to increasing values of KEY.",
        ),
        reverse=dict(
            action="store_true",
            help="reverse the order of the sort, from larger to smaller",
        ),
        **MATCHER,
    )
    def do_cache(self, args):
        """
        Cache command to inspect the CliMetLab cache.
        The selection arguments are the same as for the ``climetlab decache`` deletion
        command.
        Examples: climetlab cache --all
        """
        from climetlab.core.caching import cache_directory, dump_cache_database

        if args.path:
            print(cache_directory())
            return

        self.matcher = Matcher(args)
        cache = dump_cache_database(matcher=self.matcher)
        return self.generate_output(cache, args)

    def generate_output(self, cache, args):
        from climetlab.core.caching import cache_directory

        if not args.json and not self.matcher.undefined:
            message = humanize.list_to_human(self.matcher.message)
            print(colored(f"Entries {message}.", "green"))

        if args.sort:
            kind = None
            for e in cache:
                if e[args.sort] is not None:
                    kind = type(e[args.sort])
                    break

            if kind is not None:
                _ = {
                    dict: lambda x: tuple() if x is None else tuple(sorted(x.items())),
                    str: lambda x: "" if x is None else x,
                    int: lambda x: 0 if x is None else x,
                }[kind]

                cache = sorted(
                    cache,
                    key=lambda x: _(x[args.sort]),
                    reverse=args.reverse,
                )

        if args.json:
            print(self.as_json(cache))
            return

        if args.all:
            for entry in cache:
                print(colored(entry.pop("path"), "blue"))
                for k in (
                    "creation_date",
                    "last_access",
                    "accesses",
                    "type",
                    "size",
                    "owner",
                ):
                    print(" ", f"{k}:", colored(entry.pop(k), "green"))
                for k in sorted(entry.keys()):
                    print(" ", f"{k}:", colored(entry.pop(k), "green"))
                print()

            return

        def generate_table():
            youngest_created = None
            oldest_created = None
            youngest_accessed = None
            oldest_accessed = None
            total = 0
            for i, entry in enumerate(cache):
                if entry["size"] is not None:
                    total += entry["size"]

                if i == 0:
                    youngest_accessed = oldest_accessed = entry["last_access"]
                    youngest_created = oldest_created = entry["creation_date"]

                youngest_accessed = max(youngest_accessed, entry["last_access"])
                oldest_accessed = min(oldest_accessed, entry["last_access"])
                youngest_created = max(youngest_created, entry["last_access"])
                oldest_created = min(oldest_created, entry["last_access"])

            yield ("Cache directory:", cache_directory())
            yield ("Cache size:", humanize.bytes(total))
            yield ("Number of entries in cache:", humanize.number(len(cache)))

            if youngest_accessed:
                yield (
                    "Most recently accessed:",
                    humanize.when(datetime.datetime.fromisoformat(youngest_accessed)),
                )
                yield (
                    "Least recently accessed:",
                    humanize.when(datetime.datetime.fromisoformat(oldest_accessed)),
                )
                yield (
                    "Youngest entry:",
                    humanize.when(datetime.datetime.fromisoformat(youngest_created)),
                )
                yield (
                    "Oldest entry:",
                    humanize.when(datetime.datetime.fromisoformat(oldest_created)),
                )

        print_table(generate_table())

    def as_json(self, cache):
        return json.dumps(cache, sort_keys=True, indent=4)

    @parse_args(
        all=dict(action="store_true"),
        **MATCHER,
    )
    def do_decache(self, args):
        """
        Cache deletion command (decache) to clean the cache from selected files.
        The selection arguments are the same as for the ``climetlab cache`` query
        command.
        """

        from climetlab.core.caching import purge_cache

        matcher = Matcher(args)
        if not matcher.undefined:
            message = humanize.list_to_human(matcher.message)
            print(colored(f"Purging cache entries {message}.", "green"))
            purge_cache(matcher=matcher)
            return

        if args.all:
            purge_cache()
            return

        print(
            colored(
                "To wipe the cache completely, please use the --all flag. Use --help for more information.",
                "red",
            )
        )

    @parse_args(
        directory=(
            None,
            dict(
                type=str,
                metavar="DIRECTORY",
                help="Output directory where to copy the cache content.",
            ),
        ),
        **MATCHER,
        permissions=dict(
            type=str,
            help="""
            By default, the exported data is public and writable by its owner.
            --permissions read-only : Make the data public and read only.
            --permissions disabled : Do not tweak the files or directories permissions.
            """,
            nargs="*",
        ),
    )
    def do_export_cache(self, args):
        """
        Copy part of the cache content to a directory.
        """
        import stat

        from climetlab.core.caching import dump_cache_database

        directory = args.directory

        def permissions(perms):
            if perms is None:
                perms = []

            if perms:
                perms = [_.replace("-", "").replace("_", "") for _ in perms]

            def make_directory_writeable():
                mode = os.stat(directory).st_mode
                os.chmod(directory, stat.S_IWUSR | stat.S_IXUSR | mode)

            def check_readable():
                st = os.stat(directory)
                if not bool(st.st_mode & stat.S_IROTH):
                    LOG.warning(
                        f"The directory {directory} is not readable by others. Please make sure this is what you want."
                    )

            if "disabled" in perms:
                return (False, False, [])

            if "readonly" in perms:
                return (0o555, 0o444, [make_directory_writeable, check_readable])

            if not perms:  # default
                return (0o755, 0o644, [make_directory_writeable, check_readable])

            raise ValueError(perms)

        permissions_dirs, permissions_files, finalize = permissions(args.permissions)

        class Dirs:
            def __init__(self):
                self.new_dirs = []

            def append(self, d):
                if d in self.new_dirs:
                    return
                if not d:
                    return
                if os.path.exists(d) and os.path.isdir(d):
                    return
                parent = os.path.dirname(d)
                self.append(parent)

                os.mkdir(d)
                self.new_dirs.append(d)

            def update_permission_dirs(self, perms):
                # order may matters:
                # self.new_dirs = sorted(self.new_dirs)
                for d in self.new_dirs:
                    os.chmod(d, perms)

        new_dirs = Dirs()

        new_dirs.append(directory)

        matcher = Matcher(args)
        if not matcher.undefined:
            message = humanize.list_to_human(matcher.message)
            print(colored(f"Copying cache entries {message} to {directory}.", "green"))

        cache = dump_cache_database(matcher=matcher)

        def filter(entry):
            if "path" not in entry:
                print(f"No path. Cannot copy {entry}")
                return False
            # path = entry["path"]
            if entry.get("owner") == "grib-index":
                return False
            return True

        cache_dir = SETTINGS.get("cache-directory")

        def copy(path, dest):
            path_display = path.replace(cache_dir, "CACHE:")
            LOG.debug(f"Copying {path_display} to {dest}")

            if os.path.isdir(path):
                shutil.copytree(path, dest)
                new_dirs.append(dest)
            else:
                shutil.copy2(path, dest)
                if permissions_files:
                    os.chmod(dest, permissions_files)

        cache = [entry for entry in cache if filter(entry)]

        count = 0
        errors = 0
        for entry in tqdm(iterable=cache):
            path = entry["path"]

            assert os.path.dirname(path) == cache_dir, path
            basename = os.path.basename(path)
            dest = os.path.join(directory, basename)

            new_dirs.append(os.path.dirname(dest))

            try:
                copy(path, dest)
                count += 1
            except FileNotFoundError as e:
                LOG.exception(e)
                errors += 1

        if permissions_dirs:
            LOG.info("All entries copied. Now setting permissions.")
            new_dirs.update_permission_dirs(permissions_dirs)

        print(colored(f"Copied {count} cache entries to {directory}.", "green"))
        if errors:
            print(colored(f"Ignored {errors} errors).", "red"))

        for f in finalize:
            f()


CacheCmd.do_decache.__doc__ += "Hello"



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import subprocess

import ecmwfapi

from climetlab.core.settings import SETTINGS
from climetlab.core.temporary import temp_file

from .ecmwf_api import ECMWFApi, MARSAPIKeyPrompt

LOG = logging.getLogger(__name__)


class StandaloneMarsClient:
    EXE = "/usr/local/bin/mars"

    def execute(self, request, target):
        req = ["retrieve,"]

        for k, v in request.items():
            if isinstance(v, (list, tuple)):
                v = "/".join([str(x) for x in v])
            req += [f"{k}={v},"]

        req += [f'target="{target}"']
        req_str = "\n".join(req)
        with temp_file() as filename:
            with open(filename, "w") as f:
                f.write(req_str + "\n")
            LOG.debug(f"Sending Mars request: '{req_str}'")

            subprocess.run(
                [self.EXE, filename],
                env=dict(os.environ, MARS_AUTO_SPLIT_BY_DATES="1"),
                check=True,
            )


class MarsRetriever(ECMWFApi):
    def service(self):
        if SETTINGS.get("use-standalone-mars-client-when-available"):
            if os.path.exists(StandaloneMarsClient.EXE):
                return StandaloneMarsClient()

        prompt = MARSAPIKeyPrompt()
        prompt.check()

        try:
            return ecmwfapi.ECMWFService("mars")
        except Exception as e:
            if ".ecmwfapirc" in str(e):
                prompt.ask_user_and_save()
                return ecmwfapi.ECMWFService("mars")

            raise


source = MarsRetriever



# (C) Copyright 2020- ECMWF.  #
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from abc import abstractmethod
from collections import defaultdict

import climetlab
from climetlab.loaders import build_remapping

LOG = logging.getLogger(__name__)

PRIVATE_ATTRIBUTES = {"observer": lambda: None}


class MetaBase(type):
    def __call__(cls, *args, **kwargs):
        obj = cls.__new__(cls, *args, **kwargs)
        args, kwargs = cls.patch(obj, *args, **kwargs)
        obj.__init__(*args, **kwargs)
        return obj

    def patch(cls, obj, *args, **kwargs):
        private_attributes = {}
        private_attributes.update(PRIVATE_ATTRIBUTES)
        private_attributes.update(kwargs.pop("_PRIVATE_ATTRIBUTES", {}))

        for k, v in private_attributes.items():
            setattr(obj, k, kwargs.pop(k, v))

        return args, kwargs


class Base(metaclass=MetaBase):
    # Convertors
    def to_numpy(self, **kwargs):
        self._not_implemented()

    @abstractmethod
    def to_xarray(self, **kwargs):
        self._not_implemented()

    @abstractmethod
    def to_pandas(self, **kwargs):
        self._not_implemented()

    @abstractmethod
    def to_tfdataset(self, **kwargs):
        self._not_implemented()

    @abstractmethod
    def to_metview(self, **kwargs):
        self._not_implemented()

    # Change class
    def mutate(self):
        return self

    @classmethod
    def merge(cls, *args, **kwargs):
        return None

    # Used when plotting
    @abstractmethod
    def plot_map(self, backend):
        self._not_implemented()

    @abstractmethod
    def field_metadata(self):
        self._not_implemented()

    # I/O
    @abstractmethod
    def save(self, path):
        self._not_implemented()

    @abstractmethod
    def write(self, f):
        self._not_implemented()

    # Used by normalisers
    @abstractmethod
    def to_datetime(self):
        self._not_implemented()

    @abstractmethod
    def to_datetime_list(self):
        self._not_implemented()

    @abstractmethod
    def to_bounding_box(self):
        self._not_implemented()

    # For machine learning
    @abstractmethod
    def statistics(self):
        self._not_implemented()

    @abstractmethod
    def scaled(self, args, kwargs):
        self._not_implemented()

    @abstractmethod
    def sel(self, *args, **kwargs):
        self._not_implemented()

    @abstractmethod
    def order_by(self, *args, **kwargs):
        self._not_implemented()

    def to_pytorch_dataset(self, *args, **kwargs):
        self._not_implemented()

    def set_options(self, *args, **kwargs):
        if not hasattr(self, "_options"):
            self._options = {}
        for dic in args:
            self.set_options(**dic)
        if kwargs:
            self._options.update(**kwargs)

    def get_options(self):
        if not hasattr(self, "_options"):
            self._options = {}
        return self._options

    def cube(self, *args, **kwargs):
        from climetlab.indexing.cube import FieldCube

        return FieldCube(self, *args, **kwargs)

    def get_metadata(self, i):
        return self[i].metadata()

    def unique_values(self, *coords, remapping=None, progress_bar=True):
        """
        Given a list of metadata attributes, such as date, param, levels,
        returns the list of unique values for each attributes
        """

        assert len(coords)
        assert all(isinstance(k, str) for k in coords), coords

        remapping = build_remapping(remapping)
        iterable = self

        if progress_bar:
            iterable = climetlab.utils.progress_bar(
                iterable=self,
                desc=f"Finding coords in dataset for {coords}",
            )

        dic = defaultdict(dict)
        for f in iterable:
            metadata = remapping(f.metadata)
            for k in coords:
                v = metadata(k)
                dic[k][v] = True

        dic = {k: tuple(values.keys()) for k, values in dic.items()}

        return dic

    def combinations(self, *coords, progress_bar=True):
        assert all(isinstance(k, str) for k in coords), coords

        iterable = self

        if progress_bar:
            iterable = climetlab.utils.progress_bar(
                iterable=self,
                desc=f"Finding coords in dataset for {coords}",
            )

        for f in iterable:
            yield {k: f.metadata(k) for k in coords}

    def __add__(self, other):
        self._not_implemented()

    def _not_implemented(self):
        import inspect

        func = inspect.stack()[1][3]
        module = self.__class__.__module__
        name = self.__class__.__name__

        extra = ""
        if hasattr(self, "path"):
            extra = f" on {self.path}"
        raise NotImplementedError(f"{module}.{name}.{func}(){extra}")



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import warnings

from climetlab.decorators import normalize


class normalize_args:
    def __init__(self, **dic):
        warnings.warn(
            "Deprecated decorator @normalize_arg. Use @normalise on each argument instead."
        )
        self.decorators = []
        for name, values in dic.items():
            if isinstance(values, list):
                self.decorators.append(normalize(name, values, multiple=True))
                continue

            if isinstance(values, tuple):
                warnings.warn(
                    "Using tuple to set multiple=False is deprecated in @normalize_arg. Use with multiple=False."
                )
                self.decorators.append(normalize(name, values, multiple=False))
                continue

            self.decorators.append(normalize(name, values))

    def __call__(self, func):
        for d in self.decorators:
            func = d(func)
        return func



import code
import inspect


def get_locals(callers):
    for c in callers[1:]:
        if not c.filename.startswith("<frozen importlib"):
            return c.frame.f_locals
    import logging

    logging.debug("Cannot find calling frame.")
    return {}


frame = inspect.currentframe()
callers = inspect.getouterframes(frame, 0)

code.interact(local=get_locals(callers))



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import pathlib
import shutil
from contextlib import contextmanager
from unittest.mock import patch

from climetlab import load_source
from climetlab.readers.text import TextReader
from climetlab.sources.empty import EmptySource
from climetlab.utils import download_and_cache, module_installed

LOG = logging.getLogger(__name__)


class OfflineError(Exception):
    pass


_NETWORK_PATCHER = patch("socket.socket", side_effect=OfflineError)


@contextmanager
def network_off():
    try:
        _NETWORK_PATCHER.start()
        yield None
    finally:
        _NETWORK_PATCHER.stop()


def climetlab_file(*args):
    top = os.path.dirname(os.path.dirname(__file__))
    return os.path.join(top, *args)


def data_file(*args):
    return os.path.join(os.path.dirname(__file__), "data", *args)


def file_url(path):
    return pathlib.Path(os.path.abspath(path)).as_uri()


def data_file_url(*args):
    return file_url(data_file(*args))


def modules_installed(*modules):
    return all(module_installed(m) for m in modules)


NO_MARS = not os.path.exists(os.path.expanduser("~/.ecmwfapirc"))
NO_CDS = not os.path.exists(os.path.expanduser("~/.cdsapirc"))
IN_GITHUB = os.environ.get("GITHUB_WORKFLOW") is not None
NO_EOD = not os.path.exists(os.path.expanduser("~/.ecmwf-open-data"))


def MISSING(*modules):
    return not modules_installed(*modules)


UNSAFE_SAMPLES_URL = "https://github.com/jwilk/traversal-archives/releases/download/0"
TEST_DATA_URL = "https://get.ecmwf.int/repository/test-data/climetlab"
TEST_DATA_URL_INPUT_GRIB = TEST_DATA_URL + "/test-data/input/grib"


def empty(ds):
    LOG.debug("%s", ds)
    assert isinstance(ds, EmptySource)
    assert len(ds) == 0


def text(ds):
    LOG.debug("%s", ds)
    assert isinstance(ds._reader, TextReader)


UNSAFE_SAMPLES = (
    ("absolute1", empty),
    ("absolute2", empty),
    ("relative0", empty),
    ("relative2", empty),
    ("symlink", text),
    ("dirsymlink", text),
    ("dirsymlink2a", text),
    ("dirsymlink2b", text),
)


def check_unsafe_archives(extension):
    for archive, check in UNSAFE_SAMPLES:
        LOG.debug("%s.%s", archive, extension)
        ds = load_source("url", f"{UNSAFE_SAMPLES_URL}/{archive}{extension}")
        check(ds)


def build_testdata(dir="testdata"):
    os.makedirs(dir, exist_ok=True)
    for path in [
        "2t-tp.grib",
        "all.grib",
        "all/climetlab.json",
        "all/u.grib",
        "all/v.grib",
        "all/z.grib",
        "all/2t.grib",
        "all/climetlab.json",
        "all/tp.grib",
        "all/lsm.grib",
        "climetlab.json",
        "lsm.grib",
        "pl/climetlab.json",
        "pl/u.grib",
        "pl/v.grib",
        "pl/z.grib",
        "sfc/2t.grib",
        "sfc/climetlab.json",
        "sfc/tp.grib",
        "uvz.grib",
        "uvz-20150418.grib",
    ]:
        outpath = os.path.join(dir, path)
        if os.path.exists(outpath):
            continue
        os.makedirs(os.path.dirname(outpath), exist_ok=True)
        shutil.copyfile(
            download_and_cache(TEST_DATA_URL_INPUT_GRIB + "/" + path), outpath
        )

    return dir


@contextmanager
def cd(dir):
    old = os.getcwd()
    os.chdir(os.path.expanduser(dir))
    try:
        yield dir
    finally:
        os.chdir(old)


def main(path):
    import sys

    import pytest

    # Parallel does not work on darwin, gets RuntimeError: context has already been set
    # because pytest-parallel changes the context from `spawn` to `fork`

    args = ["-p", "no:parallel", "-E", "release"]

    if len(sys.argv) > 1 and sys.argv[1] == "--no-debug":
        args += ["-o", "log_cli=False"]
    else:
        logging.basicConfig(level=logging.DEBUG)
        args += ["-o", "log_cli=True"]

    args += [path]

    sys.exit(pytest.main(args))



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import warnings

from climetlab.indexing import PerUrlIndex
from climetlab.readers.grib.index import MultiFieldSet
from climetlab.readers.grib.index.sql import FieldsetInFilesWithSqlIndex
from climetlab.sources.indexed import IndexedSource
from climetlab.utils.patterns import Pattern


def get_index_url(url, substitute_extension, index_extension):
    if substitute_extension:
        url = ".".join(".".split(url)[:-1])

    if callable(index_extension):
        return index_extension(url)
    return url + index_extension


def add_path(url):
    def wrapped(entry):
        entry["_path"] = url
        return entry

    return wrapped


class IndexedUrls(IndexedSource):
    def __init__(
        self,
        pattern,
        request,
        substitute_extension=False,
        index_extension=".index",
        **kwargs,
    ):
        if isinstance(pattern, PerUrlIndex):
            warnings.warn(
                "Passing a PerUrlIndex object is obsolete, please update your code."
            )
            pattern = pattern.pattern
            print("PATTERN", pattern)

        pattern = Pattern(pattern, ignore_missing_keys=True)
        urls = pattern.substitute(**request)
        if not isinstance(urls, list):
            urls = [urls]

        request = dict(**request)  # deepcopy to avoid changing the user's request
        for used in pattern.names:
            # consume arguments used by Pattern to build the urls
            # This is to avoid keeping them on the request
            request.pop(used)

        index = MultiFieldSet(
            FieldsetInFilesWithSqlIndex.from_url(
                get_index_url(url, substitute_extension, index_extension),
                selection=request,
                patch_entry=add_path(url),
            )
            for url in urls
        )

        super().__init__(index, **kwargs)


source = IndexedUrls



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging

LOG = logging.getLogger(__name__)


FILEPARTS_KEY_NAMES = ["_path", "_offset", "_length"]
STATISTICS_KEY_NAMES = ["mean", "std", "min", "max", "shape"]
MORE_KEY_NAMES_WITH_UNDERSCORE = ["_param_id"]
MORE_KEY_NAMES = ["datetime", "param_level"]


class DBKey:
    cast = None

    def __init__(self, name):
        self.name = name


class StrDBKey(DBKey):
    sql_type = "TEXT"
    cast = str


class FloatDBKey(DBKey):
    sql_type = "FLOAT"
    cast = float


class IntDBKey(DBKey):
    sql_type = "INTEGER"
    cast = int


# class
# stream
# levtype
# type
# expver
# date
# hdate
# andate
# time
# antime
# reference
# ribKey(
# anoffset
# verify
# fcmonth
# fcperiod
# leadtime
# opttime
# origin
# domain
# method
# diagnostic
# iteration
# number
# quantile
# levelist
#   # in the MARS vocabulary but not used.
#
# range
# param
# ident
# obstype
# instrument
# reportype
# frequency
# direction
# channel
#
# param_level
# Key(


class Database:
    def lookup_parts(self):
        raise NotImplementedError("")

    def lookup_dicts(self, *args, **kwargs):
        raise NotImplementedError("")

    def count(self, request):
        raise NotImplementedError("")

    def load_iterator(self, iterator):
        raise NotImplementedError("")

    def sel(self, selection):
        raise NotImplementedError("")

    def order_by(self, order):
        raise NotImplementedError("")

    def already_loaded(self, path_or_url, owner):
        return False



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import threading
import time

LOCK = threading.Lock()


class Statistics:
    def __init__(self, filename=None):
        self.collect = False
        self._current = {}
        self._events = []
        self._data = []

        if filename:
            self.load_from_json(filename)

    def append(self, event):
        self._events.append(event)
        self.process_event(event)

    def process_event(self, event):
        key = event[1]

        if key == "indexed-urls":
            assert not self._current, self._current

        for k, v in event[2].items():
            if k in self._current:
                assert self._current[k] == v
            self._current[k] = v

        if key == "transfer":  # last one
            for k in ["parts", "blocks"]:
                if k in self._current:
                    v = self._current[k]
                    # create nblocks and nparts
                    self._current["n" + k] = len(v)
                    # create size_blocks and size_parts
                    self._current["size_" + k] = sum(x[1] for x in v)
                    v = [(p[0], p[1]) for p in v]
                    v = json.dumps(v)
                    self._current[k] = v

            k = "method_args"
            if k in self._current:
                v = self._current[k]
                arg1 = 0.0
                if v:
                    arg1 = v[0]
                    try:
                        arg1 = float(arg1)
                    except Exception:
                        pass
                v = str(v)
                self._current["arg1"] = arg1
                self._current[k] = v

            self._data.append(self._current)
            self._current = {}

    def to_pandas(self):
        import pandas as pd

        return pd.DataFrame(self._data)

    def write_to_json(self, filename):
        import json

        with open(filename, "w") as f:
            json.dump(
                dict(events=self._events, data=self._data),
                f,
                indent=2,
            )

    def load_from_json(self, filename):
        with open(filename, "r") as f:
            data = json.load(f)
            self.process_event(data["events"])


global STATISTICS
STATISTICS = Statistics()


def collect_statistics(on):
    assert on in [True, False]
    with LOCK:
        global STATISTICS
        STATISTICS.collect = on


def reset_statistics():
    with LOCK:
        global STATISTICS
        STATISTICS = Statistics()


def record_statistics(name, **values):
    with LOCK:
        global STATISTICS
        if STATISTICS.collect:
            STATISTICS.append((time.time(), name, values))


def retrieve_statistics():
    with LOCK:
        global STATISTICS
        stats = STATISTICS
        STATISTICS = Statistics()
        return stats



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

# WIP.

import json
import os

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

# import climetlab as cml

# Your client credentials
with open(os.path.expanduser("~/.sentinelhubrc")) as f:
    credentials = json.loads(f.read())
client_id = credentials["client_id"]
client_secret = credentials["client_secret"]

# Create a session
client = BackendApplicationClient(client_id=client_id)
oauth = OAuth2Session(client=client)


# Get token for the session
token = oauth.fetch_token(
    token_url="https://services.sentinel-hub.com/oauth/token",
    client_id=client_id,
    client_secret=client_secret,
)

# All requests using this session will have an access token automatically added
resp = oauth.get("https://services.sentinel-hub.com/oauth/tokeninfo")
print(resp.content)



#!/usr/bin/env python3
# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

TOP = os.path.dirname(os.path.dirname(__file__))

LANGUAGES = {".py": "python", ".yaml": "yaml"}


def execute(path):
    _, ext = os.path.splitext(path)

    print()
    print(".. code-block::", LANGUAGES[ext])
    print()

    with open(os.path.join(TOP, path)) as f:
        for line in f:
            print("    ", line.rstrip())

    print()
    print()


if __name__ == "__main__":
    execute("data/projections/global.yaml")



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging

from .argument import Argument
from .transformers import AvailabilityChecker, KwargsAliasTransformer

LOG = logging.getLogger(__name__)


class InputManager:
    def __init__(
        self,
        decorators,
    ):
        self.decorators = decorators
        self._pipeline = None
        self.alias_argument = None
        self.alias_arguments = []
        self.availabilities = []
        self.arguments = {}

    @property
    def pipeline(self):
        if self._pipeline is None:
            self.build_pipeline()
        return self._pipeline

    def register_availability(self, decorator):
        for name, values in decorator.availability.unique_values().items():
            if name not in self.arguments:
                self.arguments[name] = Argument(name)
            self.arguments[name].availability = values
        self.availabilities.append(decorator.availability)

    def register_normalize(self, decorator):
        if decorator.name not in self.arguments:
            self.arguments[decorator.name] = Argument(decorator.name)
        self.arguments[decorator.name].normalize = decorator.kwargs

    def register_alias_argument(self, decorator):
        self.alias_arguments.append(decorator)

    def build_pipeline(self):
        self._pipeline = []
        LOG.debug("Building arguments from decorators:\n %s", self.decorators)

        self.arguments = {}

        for decorator in self.decorators:
            decorator.register(self)

        self.arguments = list(self.arguments.values())

        for a in self.alias_arguments:
            transform = KwargsAliasTransformer(a)
            self._pipeline.append(transform)

        for a in self.arguments:
            a.add_alias_transformers(self._pipeline)

        for a in self.arguments:
            a.add_type_transformers(self._pipeline)

        for a in self.arguments:
            a.add_alias_transformers(self._pipeline)

        for availability in self.availabilities:
            transform = AvailabilityChecker(availability)
            self._pipeline.append(transform)

        for a in self.arguments:
            a.add_format_transformers(self._pipeline)

    def apply_to_kwargs_before_default(self, kwargs):
        LOG.debug(
            f"Apply pipeline to kwargs before resolving default values: {safe_to_str(kwargs)}"
        )
        for t in self.pipeline:
            if hasattr(t, "name"):
                LOG.debug(f" - {t.name}: apply {t}.")
            else:
                LOG.debug(f" - apply {t}.")

            kwargs = t.execute_before_default(kwargs)
            LOG.debug(f"       kwargs: {safe_to_str(kwargs)}")

        return kwargs

    def apply_to_kwargs(self, kwargs):
        LOG.debug(f"Apply pipeline to kwargs: {safe_to_str(kwargs)}")
        for t in self.pipeline:
            if hasattr(t, "name"):
                LOG.debug(f" - {t.name}: apply {t}.")
            else:
                LOG.debug(f" - apply {t}.")

            kwargs = t.execute(kwargs)
            LOG.debug(f"       kwargs: {safe_to_str(kwargs)}")

        return kwargs

    def apply_to_value(self, value):
        for t in self.pipeline:
            value = t.__call__(value)
        return value

    def __repr__(self) -> str:
        txt = "ARGUMENTS:[\n"
        txt += " availability >>>\n"
        if self.availabilities:
            txt += f"{self.availabilities}<<<"
        else:
            txt += "None<<<"
        txt += "\n"
        for a in self.arguments:
            txt += f"  {a}\n"
        txt += "]"
        if self._pipeline is None:
            txt += "Pipeline[not-ready]"
        else:
            txt += "Pipeline:[\n"
            for t in self._pipeline:
                txt += f"  {t}\n"
            txt += "]"
        return txt

    def get_argument(self, name):
        for a in self.arguments:
            if a == name:
                return a
        return None

    def consolidate_defaults(self, defaults):
        for k, default in defaults.items():
            a = self.get_argument(k)
            if a:
                a.set_default(default)

    def apply_to_arg_kwargs(self, args, kwargs, func):
        from climetlab.arguments.args_kwargs import ArgsKwargs

        args_kwargs = ArgsKwargs(args, kwargs, func=func)
        LOG.debug("Applying decorator stack (before default) to: %s %s", args, kwargs)
        args_kwargs.kwargs = self.apply_to_kwargs_before_default(args_kwargs.kwargs)

        args_kwargs.add_default_values_and_kwargs()
        self.consolidate_defaults(args_kwargs.defaults)

        LOG.debug("Applying decorator stack to: %s %s", args, kwargs)
        args_kwargs.kwargs = self.apply_to_kwargs(args_kwargs.kwargs)

        args_kwargs.ensure_positionals_only()

        args, kwargs = args_kwargs.args, args_kwargs.kwargs

        LOG.debug("CALLING func %s %s", args, kwargs)
        return args, kwargs


def normaliser(*args, **kwargs):
    return Argument("no-name", *args, **kwargs)


def safe_to_str(dic):
    lst = []
    for k, v in dic.items():
        try:
            v = str(v)
        except:  # noqa: E722
            v = "..."
        lst.append(f"{k}={v}")
    return ",".join(lst)



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import tensorflow as tf
from tensorflow.keras.layers import Conv2D

LOG = logging.getLogger(__name__)


class PeriodicConv2D(Conv2D):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.kernel_size

    def call(self, inputs):
        w, h = self.kernel_size
        inputs = tf.concat([inputs, inputs[:, :, :w, :]], axis=2)
        inputs = tf.pad(
            inputs, [[0, 0], [h // 2, h // 2], [0, 0], [0, 0]], constant_values=0
        )
        return super().call(inputs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import itertools
import logging

from climetlab.core.thread import SoftThreadPool
from climetlab.mergers import make_merger, merge_by_class
from climetlab.sources.empty import EmptySource
from climetlab.utils import tqdm
from climetlab.utils.bbox import BoundingBox

from . import Source

LOG = logging.getLogger(__name__)


class MultiSource(Source):
    def __init__(self, *sources, filter=None, merger=None):
        if len(sources) == 1 and isinstance(sources[0], list):
            sources = sources[0]

        sources = self._load_sources(sources)

        self.sources = [s.mutate() for s in sources if not s.ignore()]
        self.filter = filter
        self.merger = merger
        self._lengths = [None] * len(self.sources)

    def ignore(self):
        return len(self.sources) == 0

    def mutate(self):
        if len(self.sources) == 1:
            return self.sources[0].mutate()

        if len(self.sources) == 0:
            return EmptySource()

        if self.merger is None:
            merged = merge_by_class(self.sources)
            if merged is not None:
                return merged.mutate()

        return self

    def _set_dataset(self, dataset):
        super()._set_dataset(dataset)
        for s in self.sources:
            s._set_dataset(dataset)

    def __iter__(self):
        return itertools.chain(*self.sources)

    def __getitem__(self, n):
        if n < 0:
            n = len(self) + n

        i = 0
        while n >= self._length(i):
            n -= self._length(i)
            i += 1
        return self.sources[i][n]

    def sel(self, *args, **kwargs):
        new_sources = [s.sel(*args, **kwargs) for s in self.sources]
        return self.__class__(new_sources, filter=self.filter, merger=self.merger)

    def order_by(self, *args, **kwargs):
        new_sources = [s.order_by(*args, **kwargs) for s in self.sources]
        return self.__class__(new_sources, filter=self.filter, merger=self.merger)

    def __len__(self):
        return sum(self._length(i) for i, _ in enumerate(self.sources))

    def _length(self, i):
        if self._lengths[i] is None:
            self._lengths[i] = len(self.sources[i])
        return self._lengths[i]

    def __repr__(self) -> str:
        string = ",".join(repr(s) for s in self.sources)
        return f"{self.__class__.__name__}({string})"

    def save(self, path):
        with open(path, "wb") as f:
            for s in self.sources:
                s.write(f)

    def graph(self, depth=0):
        print(" " * depth, self.__class__.__name__, self.merger)
        for s in self.sources:
            s.graph(depth + 3)

    def to_xarray(self, **kwargs):
        return make_merger(self.merger, self.sources).to_xarray(**kwargs)

    def to_tfdataset(self, **kwargs):
        return make_merger(self.merger, self.sources).to_tfdataset(**kwargs)

    def to_pandas(self, **kwargs):
        return make_merger(self.merger, self.sources).to_pandas(**kwargs)

    def statistics(self, **kwargs):
        return make_merger(self.merger, self.sources).statistics(**kwargs)

    def _load_sources(self, sources):
        callables = []
        has_callables = False
        for s in sources:
            if callable(s):
                has_callables = True
                callables.append(s)
            else:
                callables.append(lambda *args, **kwargs: s)

        if not has_callables:
            return sources

        nthreads = min(self.settings("number-of-download-threads"), len(callables))
        if nthreads < 2:
            return [s() for s in sources]

        def _call(s, *args, **kwargs):
            return s(*args, **kwargs)

        with SoftThreadPool(nthreads=nthreads) as pool:
            futures = [pool.submit(_call, s, observer=pool) for s in sources]

            iterator = (f.result() for f in futures)
            sources = list(tqdm(iterator, leave=False, total=len(futures)))

        return sources

    # Used by normalisers
    def to_datetime(self):
        times = self.to_datetime_list()
        assert len(times) == 1
        return times[0]

    def to_datetime_list(self):
        result = set()
        for s in self.sources:
            result.update(s.to_datetime_list())
        return sorted(result)

    def to_bounding_box(self):
        return BoundingBox.multi_merge([s.to_bounding_box() for s in self.sources])

    def plot_map(self, *args, **kwargs):
        # TODO: we plot the first one for now
        return self.sources[0].plot_map(*args, **kwargs)


source = MultiSource



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import weakref

ANNOTATIONS = []


def free_slot():
    for i, a in enumerate(ANNOTATIONS):
        if a is None:
            return i
        if a.owner is None:
            return i
    ANNOTATIONS.append(None)
    return len(ANNOTATIONS) - 1


class Annotation:
    def __init__(self, owner, **kwargs):
        self._owner = None if owner is None else weakref.ref(owner)
        self._kwargs = kwargs

    def get(self, name):
        return self._kwargs.get(name)

    @property
    def owner(self):
        if self._owner is None:
            return None
        return self._owner()

    def __repr__(self):
        return repr(self._kwargs)


def _annotate_pandas(pd, owner, **kargs):
    n = None
    for i, a in enumerate(pd._metadata):
        if isinstance(a, str) and a.startswith("climetlab-"):
            n = i

    if n is None:
        n = len(pd._metadata)
        pd._metadata.append(None)

    slot = free_slot()
    pd._metadata[n] = "climetlab-{}".format(slot)

    ANNOTATIONS[slot] = Annotation(owner, **kargs)


def _annotation_pandas(pd):
    for a in pd._metadata:
        if isinstance(a, str) and a.startswith("climetlab-"):
            _, i = a.split("-")
            return ANNOTATIONS[int(i)]

    return Annotation(None)


def _annotate_xarray(xr, owner, **kargs):
    xr.climetlab._metadata = Annotation(owner, **kargs)


def _annotation_xarray(xr):
    return xr.climetlab._metadata


def _annotation_numpy(array):
    # TODO
    # See https://numpy.org/doc/stable/user/basics.subclassing.html
    pass


def annotate(obj, owner, **kwargs):
    if hasattr(obj, "_metadata"):
        _annotate_pandas(obj, owner, **kwargs)
        return obj

    if hasattr(obj, "climetlab"):
        _annotate_xarray(obj, owner, **kwargs)
        return obj

    raise NotImplementedError("Cannot annotate object of type", type(obj))


def annotation(obj):
    if hasattr(obj, "_metadata"):
        return _annotation_pandas(obj)

    if hasattr(obj, "climetlab"):
        return _annotation_xarray(obj)

    raise NotImplementedError("Cannot annotate object of type", type(obj))


class XMetadata:
    def __init__(self, xarray_obj):
        # self._obj = xarray_obj
        self._metadata = Annotation(None)


class XDataset(XMetadata):
    def __init__(self, xarray_obj):
        super().__init__(xarray_obj)


class XDataArray(XMetadata):
    def __init__(self, xarray_obj):
        super().__init__(xarray_obj)


if False:
    # This takes too long
    try:
        import xarray as xr

        xr.register_dataset_accessor("climetlab")(XDataset)
        xr.register_dataarray_accessor("climetlab")(XDataArray)
    except Exception:
        pass


def init_metadata():
    # Dummy function so climetlab.__init__ loads that file
    # and the xarray accessors are registered
    pass



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pandas as pd


def merge(
    sources=None,
    paths=None,
    reader_class=None,
    **kwargs,
):
    options = dict(ignore_index=True)  # Renumber all indices
    options.update(kwargs)
    return pd.concat([s.to_pandas(**kwargs) for s in sources], **options)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import ecmwfapi

from .ecmwf_api import ECMWFApi


class Wrapper(ecmwfapi.ECMWFDataServer):
    def __init__(self, dataset):
        super().__init__()
        self.dataset = dataset

    def execute(self, request, target):
        request = dict(**request)
        request["dataset"] = self.dataset
        request["target"] = target
        print(request)
        return self.retrieve(request)


class DatasetRetrieverBase(ECMWFApi):
    def __init__(self, dataset, *args, **kwargs):
        self.ecmwf_dataset = dataset
        super().__init__(*args, **kwargs)

    def service(self):
        return Wrapper(self.ecmwf_dataset)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import itertools

import numpy as np

from climetlab.utils.dates import to_datetime_list
from climetlab.wrappers import Wrapper


def find_lat_lon(data, variable=None):
    latitude = None
    longitude = None

    LATITUDES = [
        lambda name, v: v.attrs.get("standard_name") == "latitude",
        # lambda name, v: v.attrs.get("standard_name") == "projection_x_coordinate",
        lambda name, v: v.attrs.get("long_name") == "latitude",
        lambda name, v: v.attrs.get("long_name") == "projection_x_coordinate",
        lambda name, v: name.lower() in ["latitude", "lat"],
    ]

    LONGITUDES = [
        lambda name, v: v.attrs.get("standard_name") == "longitude",
        # lambda name, v: v.attrs.get("standard_name") == "projection_y_coordinate",
        lambda name, v: v.attrs.get("long_name") == "longitude",
        lambda name, v: v.attrs.get("long_name") == "projection_y_coordinate",
        lambda name, v: name.lower() in ["longitude", "lon"],
    ]

    for trigger in LATITUDES:
        for name, v in itertools.chain(data.coords.items(), data.data_vars.items()):
            if latitude is None and trigger(name, v):
                latitude = v

    for trigger in LONGITUDES:
        for name, v in itertools.chain(data.coords.items(), data.data_vars.items()):
            if longitude is None and trigger(name, v):
                longitude = v

    # Else, use latest coordinates from the variable
    if latitude is None or longitude is None and variable is not None:
        assert latitude is None and longitude is None, (latitude, longitude)

        lat, lon = variable.dims[-2], variable.dims[-1]
        latitude = data[lat]
        longitude = data[lon]

    assert latitude.name != longitude.name, latitude.name

    return latitude, longitude


class XArrayDatasetWrapper(Wrapper):
    def __init__(self, data):
        self.data = data
        dims = 0
        for name, var in data.data_vars.items():
            # Choose variable with the most dimensions
            if len(var.dims) > dims:
                self.name = name
                self.var = var
                dims = len(var.dims)

        self.latitude, self.longitude = find_lat_lon(self.data, self.var)

        # For Magics
        self.latitude.attrs["standard_name"] = "latitude"
        self.longitude.attrs["standard_name"] = "longitude"

        self.north = np.amax(self.latitude.data)
        self.south = np.amin(self.latitude.data)
        self.east = np.amax(self.longitude.data)
        self.west = np.amin(self.longitude.data)

    def plot_map(self, backend):
        backend.bounding_box(
            north=self.north, south=self.south, west=self.west, east=self.east
        )

        dimension_settings = dict()

        extra_dims = list(self.var.dims[:])
        assert self.latitude.name in extra_dims, (self.latitude.name, extra_dims)
        extra_dims.remove(self.latitude.name)
        assert self.longitude.name in extra_dims, (self.longitude.name, extra_dims)
        extra_dims.remove(self.longitude.name)

        for d in extra_dims:
            dimension_settings[d] = 0  # self.data[d].data[0]

        assert self.data[self.latitude.name].attrs["standard_name"] == "latitude"
        assert self.data[self.longitude.name].attrs["standard_name"] == "longitude"

        backend.plot_xarray(self.data, self.name, dimension_settings)

    def field_metadata(self):
        shape = self.var.shape
        result = dict(shape=(shape[-2], shape[-1]))
        result.update(self.var.attrs)
        result.update(
            dict(
                north=self.north,
                south=self.south,
                east=self.east,
                west=self.west,
                # # TODO:
                # south_north_increment=None,
                # west_east_increment=None,
            )
        )
        return result

    def to_xarray(self):
        return self.data


class XArrayDataArrayWrapper(Wrapper):
    def __init__(self, data):
        self.data = data

    def to_datetime_list(self):
        return to_datetime_list(self.data.values)


def wrapper(data, *args, **kwargs):
    import xarray as xr

    if isinstance(data, xr.Dataset):
        return XArrayDatasetWrapper(data, *args, **kwargs)

    if isinstance(data, xr.DataArray):
        try:
            return XArrayDatasetWrapper(data.to_dataset(), *args, **kwargs)
        except ValueError:
            return XArrayDataArrayWrapper(data, *args, **kwargs)

    return None



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import datetime

from climetlab.wrappers import Wrapper, get_wrapper


class NumpyArrayWrapper(Wrapper):
    def __init__(self, data, *args, **kwargs):
        self.data = data

    def plot_map(self, backend):
        wrapper = get_wrapper(backend.option("metadata"))

        if hasattr(wrapper, "plot_numpy"):
            return wrapper.plot_numpy(backend, self.data)

        metadata = wrapper.field_metadata()

        backend.bounding_box(
            north=metadata["north"],
            south=metadata["south"],
            west=metadata["west"],
            east=metadata["east"],
        )

        backend.plot_numpy(
            self.data.reshape(metadata.get("shape", self.data.shape)),
            metadata=metadata,
        )

    def to_datetime_list(self):
        return [
            datetime.datetime.fromtimestamp(x * 1e-9, tz=datetime.timezone.utc)
            for x in self.data.tolist()
        ]


def wrapper(data, *args, **kwargs):
    import numpy as np

    if isinstance(data, np.ndarray):
        return NumpyArrayWrapper(data, *args, **kwargs)
    return None



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging

from multiurl import Downloader

from climetlab.core.caching import cache_file
from climetlab.core.settings import SETTINGS
from climetlab.core.statistics import record_statistics
from climetlab.utils import progress_bar

from .file import FileSource

LOG = logging.getLogger(__name__)


def download_and_cache(
    url,
    *,
    owner="url",
    parts=None,
    verify=True,
    force=None,
    chunk_size=1024 * 1024,
    range_method="auto",
    http_headers=None,
    update_if_out_of_date=False,
    fake_headers=None,  # When HEAD is not allowed but you know the size
    **kwargs,
):
    # TODO: re-enable this feature
    extension = None

    LOG.debug("URL %s", url)

    downloader = Downloader(
        url,
        chunk_size=chunk_size,
        timeout=SETTINGS.get("url-download-timeout"),
        verify=verify,
        parts=parts,
        range_method=range_method,
        http_headers=http_headers,
        fake_headers=fake_headers,
        statistics_gatherer=record_statistics,
        progress_bar=progress_bar,
        resume_transfers=True,
        override_target_file=False,
        download_file_extension=".download",
    )

    if extension and extension[0] != ".":
        extension = "." + extension

    if extension is None:
        extension = downloader.extension()

    path = downloader.local_path()
    if path is not None:
        return

    def out_of_date(url, path, cache_data):
        if SETTINGS.get("check-out-of-date-urls") is False:
            return False

        if downloader.out_of_date(path, cache_data):
            if SETTINGS.get("download-out-of-date-urls") or update_if_out_of_date:
                LOG.warning(
                    "Invalidating cache version and re-downloading %s",
                    url,
                )
                return True
            else:
                LOG.warning(
                    "To enable automatic downloading of updated URLs set the 'download-out-of-date-urls'"
                    " setting to True",
                )
        return False

    if force is None:
        force = out_of_date

    def download(target, _):
        downloader.download(target)
        return downloader.cache_data()

    path = cache_file(
        owner,
        download,
        dict(url=url, parts=parts),
        extension=extension,
        force=force,
    )

    return path


class Url(FileSource):
    def __init__(
        self,
        url,
        parts=None,
        filter=None,
        merger=None,
        verify=True,
        force=None,
        chunk_size=1024 * 1024,
        range_method="auto",
        http_headers=None,
        update_if_out_of_date=False,
        fake_headers=None,  # When HEAD is not allowed but you know the size
    ):
        super().__init__(filter=filter, merger=merger)

        # TODO: re-enable this feature
        extension = None

        self.url = url
        self.parts = parts
        LOG.debug("URL %s", url)

        self.update_if_out_of_date = update_if_out_of_date

        self.downloader = Downloader(
            url,
            chunk_size=chunk_size,
            timeout=SETTINGS.get("url-download-timeout"),
            verify=verify,
            parts=parts,
            range_method=range_method,
            http_headers=http_headers,
            fake_headers=fake_headers,
            statistics_gatherer=record_statistics,
            progress_bar=progress_bar,
            resume_transfers=True,
            override_target_file=False,
            download_file_extension=".download",
        )

        if extension and extension[0] != ".":
            extension = "." + extension

        if extension is None:
            extension = self.downloader.extension()

        self.path = self.downloader.local_path()
        if self.path is not None:
            return

        if force is None:
            force = self.out_of_date

        def download(target, _):
            self.downloader.download(target)
            return self.downloader.cache_data()

        self.path = self.cache_file(
            download,
            dict(url=url, parts=parts),
            extension=extension,
            force=force,
        )

    def connect_to_mirror(self, mirror):
        return mirror.connection_for_url(self, self.url, self.parts)

    def out_of_date(self, url, path, cache_data):
        if SETTINGS.get("check-out-of-date-urls") is False:
            return False

        if self.downloader.out_of_date(path, cache_data):
            if SETTINGS.get("download-out-of-date-urls") or self.update_if_out_of_date:
                LOG.warning(
                    "Invalidating cache version and re-downloading %s",
                    self.url,
                )
                return True
            else:
                LOG.warning(
                    "To enable automatic downloading of updated URLs set the 'download-out-of-date-urls'"
                    " setting to True",
                )
        return False

    def __repr__(self) -> str:
        return f"Url({self.url})"


source = Url



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

"""

Internally, CliMetLab cache is managed by the module `climetlab.core.cache`,
it relies on a sqlite database. The :py:func:`cache_file` function provide
a unique path for a given couple (`owner`, `args`).
The calling code is responsible for checking if the file exists and
decide to read it or create it.

"""

import ctypes
import datetime
import functools
import hashlib
import json
import logging
import os
import platform
import shutil
import sqlite3
import threading
import time

from filelock import FileLock

from climetlab.core.settings import SETTINGS
from climetlab.utils import humanize
from climetlab.utils.html import css

VERSION = 2
CACHE_DB = f"cache-{VERSION}.db"

LOG = logging.getLogger(__name__)


CONNECTION = None
CACHE = None


class DiskUsage:
    def __init__(self, path):
        path = os.path.realpath(path)
        self.path = path

        if platform.system() == "Windows":
            avail = ctypes.c_ulonglong()
            total = ctypes.c_ulonglong()
            free = ctypes.c_ulonglong()

            ctypes.windll.kernel32.GetDiskFreeSpaceExW(
                ctypes.c_wchar_p(path),
                ctypes.pointer(avail),
                ctypes.pointer(total),
                ctypes.pointer(free),
            )
            self.avail = avail.value
            self.total = total.value
            self.free = free.value
        else:
            st = os.statvfs(path)
            self.free = st.f_bfree * st.f_frsize
            self.total = st.f_blocks * st.f_frsize
            self.avail = st.f_bavail * st.f_frsize

        self.percent = int(
            float(self.total - self.avail) / float(self.total) * 100 + 0.5
        )

    def __repr__(self):
        return (
            f"DiskUsage(total={self.total},free={self.free},"
            f"avail={self.avail},percent={self.percent},path={self.path})"
        )


def disk_usage(path):
    return DiskUsage(path)


def default_serialiser(o):
    import pandas as pd

    if isinstance(o, (datetime.date, datetime.datetime)):
        return o.isoformat()
    if isinstance(o, (pd.Timestamp)):
        return o.isoformat()
    if isinstance(o, (pd.DatetimeIndex)):
        return [_.isoformat() for _ in o]
    return json.JSONEncoder.default(o)


def in_executor(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        global CACHE
        s = CACHE.enqueue(func, *args, **kwargs)
        return s.result()

    return wrapped


def in_executor_forget(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        global CACHE
        CACHE.enqueue(func, *args, **kwargs)
        return None

    return wrapped


class Future:
    def __init__(self, func, args, kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self._condition = threading.Condition()
        self._ready = False
        self._result = None

    def execute(self):
        try:
            self._result = self.func(*self.args, **self.kwargs)
        except Exception as e:
            LOG.error(e)
            self._result = e
        with self._condition:
            self._ready = True
            self._condition.notify_all()

    def result(self):
        with self._condition:
            while not self._ready:
                self._condition.wait()
        if isinstance(self._result, Exception):
            raise self._result
        return self._result


class Cache(threading.Thread):
    def __init__(self):
        super().__init__(daemon=True)
        self._connection = None
        self._queue = []
        self._condition = threading.Condition()

    def run(self):
        while True:
            with self._condition:
                while len(self._queue) == 0:
                    self._condition.wait()
                s = self._queue.pop(0)
                self._condition.notify_all()
            s.execute()

    @property
    def connection(self):
        if self._connection is None:
            self._connection = self.new_connection()

        return self._connection

    def new_connection(self):
        cache_dir = SETTINGS.get("cache-directory")
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir, exist_ok=True)
        cache_db = os.path.join(cache_dir, CACHE_DB)
        LOG.debug("Cache database is %s", cache_db)
        connection = sqlite3.connect(cache_db)
        # So we can use rows as dictionaries
        connection.row_factory = sqlite3.Row

        # If you change the schema, change VERSION above
        connection.execute(
            """
            CREATE TABLE IF NOT EXISTS cache (
                    path          TEXT PRIMARY KEY,
                    owner         TEXT NOT NULL,
                    args          TEXT NOT NULL,
                    creation_date TEXT NOT NULL,
                    flags         INTEGER DEFAULT 0,
                    owner_data    TEXT,
                    last_access   TEXT NOT NULL,
                    type          TEXT,
                    parent        TEXT,
                    replaced      TEXT,
                    extra         TEXT,
                    expires       INTEGER,
                    accesses      INTEGER,
                    size          INTEGER);"""
        )
        return connection

    def enqueue(self, func, *args, **kwargs):
        with self._condition:
            s = Future(func, args, kwargs)
            self._queue.append(s)
            self._condition.notify_all()
            return s

    def _file_in_cache_directory(self, path):
        cache_directory = SETTINGS.get("cache-directory")
        return path.startswith(cache_directory)

    def _cache_directory(self):
        cache_directory = SETTINGS.get("cache-directory")
        return cache_directory

    def _ensure_in_cache(self, path):
        assert self._file_in_cache_directory(path), f"File not in cache {path}"

    def _settings_changed(self):
        LOG.debug("Settings changed")
        self._connection = None  # The user may have changed the cache directory
        self._check_cache_size()

    def _latest_date(self):
        """Returns the latest date to be used when purging the cache.
        So we do not purge files being downloaded."""
        with self.connection as db:
            latest = db.execute(
                "SELECT MIN(creation_date) FROM cache WHERE size IS NULL"
            ).fetchone()[0]
            if latest is None:
                latest = db.execute(
                    "SELECT MAX(creation_date) FROM cache WHERE size IS NOT NULL"
                ).fetchone()[0]
            if latest is None:
                latest = datetime.datetime.now()
            if isinstance(latest, str):
                latest = datetime.datetime.fromisoformat(latest)
            return latest

    def _purge_cache(self, matcher=None):
        if matcher is None:
            self._housekeeping(clean=True)
            # _update_cache(clean=True)
            self._decache(self._cache_size(), purge=True)
            return

        dump = self._dump_cache_database(matcher)
        for entry in dump:
            self._delete_entry(entry)

    def _cache_entries(self):
        result = []
        with self.connection as db:
            for n in db.execute("SELECT * FROM cache").fetchall():
                n = dict(n)
                n["args"] = json.loads(n["args"])
                try:
                    n["owner_data"] = json.loads(n["owner_data"])
                except Exception:
                    pass
                if os.path.exists(n["path"]):
                    result.append(n)
        return result

    def _update_entry(self, path, owner_data=None):
        self._ensure_in_cache(path)

        if os.path.isdir(path):
            kind = "directory"
            size = 0
            for root, _, files in os.walk(path):
                for f in files:
                    size += os.path.getsize(os.path.join(root, f))
        else:
            kind = "file"
            size = os.path.getsize(path)

        with self.connection as db:
            db.execute(
                "UPDATE cache SET size=?, type=?, owner_data=? WHERE path=?",
                (
                    size,
                    kind,
                    json.dumps(owner_data, default=default_serialiser),
                    path,
                ),
            )

    def _update_cache(self, clean=False):
        """Update cache size and size of each file in the database ."""
        with self.connection as db:
            update = []
            commit = False
            for n in db.execute("SELECT path FROM cache WHERE size IS NULL"):
                try:
                    path = n[0]
                    if os.path.isdir(path):
                        kind = "directory"
                        size = 0
                        for root, _, files in os.walk(path):
                            for f in files:
                                size += os.path.getsize(os.path.join(root, f))
                    else:
                        kind = "file"
                        size = os.path.getsize(path)
                    update.append((size, kind, path))
                except Exception:
                    if clean:
                        db.execute("DELETE from cache WHERE path=?", (path,))
                        commit = True

            if update:
                db.executemany("UPDATE cache SET size=?, type=? WHERE path=?", update)

            if update or commit:
                db.commit()

    def _housekeeping(self, clean=False):
        top = SETTINGS.get("cache-directory")
        with self.connection as db:
            for name in os.listdir(top):
                if name == CACHE_DB:
                    continue

                full = os.path.join(top, name)
                count = db.execute(
                    "SELECT count(*) FROM cache WHERE path=?", (full,)
                ).fetchone()[0]

                if count > 0:
                    continue

                parent = None
                start = full.split(".")[0] + "%"
                for n in db.execute(
                    "SELECT path FROM cache WHERE parent IS NULL and path LIKE ?",
                    (start,),
                ).fetchall():
                    if full.startswith(n["path"]):
                        parent = n["path"]
                        break

                try:
                    s = os.stat(full)
                    if time.time() - s.st_mtime < 120:  # Two minutes
                        continue
                except OSError:
                    pass

                if parent is None:
                    LOG.warning(f"CliMetLab cache: orphan found: {full}")
                else:
                    LOG.debug(
                        f"CliMetLab cache: orphan found: {full} with parent {parent}"
                    )

                self._register_cache_file(
                    full,
                    "orphans",
                    None,
                    parent,
                )
        self._update_cache(clean=clean)

    def _delete_file(self, path):
        self._ensure_in_cache(path)

        try:
            if os.path.isdir(path) and not os.path.islink(path):
                shutil.rmtree(path)
            else:
                os.unlink(path)
        except Exception as e:
            print(e)
            LOG.exception("Deleting %s", path)

    def _entry_to_dict(self, entry):
        n = dict(entry)
        for k in ("args", "owner_data"):
            if k in n and isinstance(n[k], str):
                try:
                    n[k] = json.loads(n[k])
                except Exception:
                    LOG.debug("Cannot decode JSON %s", n[k])
                    pass
        return n

    def _delete_entry(self, entry):
        if isinstance(entry, str):
            entry = dict(
                path=entry,
                size=0,
                owner=None,
                args=None,
            )
            try:
                entry["size"] = os.path.getsize(entry["path"])
            except OSError:
                pass

        # Detach from sqlite
        entry = dict(**entry)

        if entry["size"] is None:
            entry["size"] = 0

        path, size, owner, args = (
            entry["path"],
            entry["size"],
            entry["owner"],
            entry["args"],
        )

        LOG.warning(
            "Deleting entry %s",
            json.dumps(
                self._entry_to_dict(entry), indent=4, default=default_serialiser
            ),
        )
        total = 0

        # First, delete child files, e.g. unzipped data
        with self.connection as db:
            for child in db.execute("SELECT * FROM cache WHERE parent = ?", (path,)):
                total += self._delete_entry(child)

        if not os.path.exists(path):
            LOG.warning(f"cache file lost: {path}")
            with self.connection as db:
                db.execute("DELETE FROM cache WHERE path=?", (path,))
            return total

        try:
            LOG.warning(
                f"CliMetLab cache: deleting {path} ({humanize.bytes(size)}, "
                f"{humanize.seconds(time.time()-os.path.getmtime(path))})"
            )
            LOG.warning(f"CliMetLab cache: {owner} {args}")
        except OSError as e:
            LOG.exception(e)

        self._delete_file(path)

        with self.connection as db:
            db.execute("DELETE FROM cache WHERE path=?", (path,))

        return total + size

    def _decache(self, bytes, purge=False):
        # _find_orphans()
        # _update_cache(clean=True)

        if bytes <= 0:
            return 0

        LOG.warning("CliMetLab cache: trying to free %s", humanize.bytes(bytes))

        total = 0

        with self.connection as db:
            latest = datetime.datetime.now() if purge else self._latest_date()
            age = datetime.datetime.now() - latest
            age = age.days * 24 * 3600 + age.seconds
            LOG.warning(
                f"Decaching files oldest than {latest.isoformat()} (age: {humanize.seconds(age)})"
            )

            for stmt in (
                "SELECT * FROM cache WHERE size IS NOT NULL AND owner='orphans' AND creation_date < ?",
                "SELECT * FROM cache WHERE size IS NOT NULL AND creation_date < ? ORDER BY last_access ASC",
            ):
                for entry in db.execute(stmt, (latest,)):
                    total += self._delete_entry(entry)
                    if total >= bytes:
                        LOG.warning(
                            "CliMetLab cache: freed %s from cache",
                            humanize.bytes(bytes),
                        )
                        return total

        LOG.warning("CliMetLab cache: could not free %s", humanize.bytes(bytes))

    def _register_cache_file(self, path, owner, args, parent=None):
        """Register a file in the cache

        Parameters
        ----------
        path : str
            Cache file to register
        owner : str
            Owner of the cache file (generally a source or a dataset)
        args : dict
            Dictionary to save with the file in the database, as json string.

        Returns
        -------
        changes :
            None or False if database does not need to be updated. TODO: clarify.
        """

        self._ensure_in_cache(path)

        with self.connection as db:
            now = datetime.datetime.now()

            args = json.dumps(args, default=default_serialiser)

            db.execute(
                """
                UPDATE cache
                SET accesses    = accesses + 1,
                    last_access = ?
                WHERE path=?""",
                (now, path),
            )

            changes = db.execute("SELECT changes()").fetchone()[0]

            if not changes:
                db.execute(
                    """
                    INSERT INTO cache(
                                    path,
                                    owner,
                                    args,
                                    creation_date,
                                    last_access,
                                    accesses,
                                    parent)
                    VALUES(?,?,?,?,?,?,?)""",
                    (path, owner, args, now, now, 1, parent),
                )

            return dict(
                db.execute("SELECT * FROM cache WHERE path=?", (path,)).fetchone()
            )

    def _cache_size(self):
        with self.connection as db:
            size = db.execute("SELECT SUM(size) FROM cache").fetchone()[0]
            if size is None:
                size = 0
            return size

    def _decache_file(self, path):
        self._delete_entry(path)

    def _check_cache_size(self):
        # Check absolute limit
        size = self._cache_size()
        maximum = SETTINGS.get("maximum-cache-size")
        if maximum is not None and size > maximum:
            self._housekeeping()
            self._decache(size - maximum)

        # Check relative limit
        size = self._cache_size()
        usage = SETTINGS.get("maximum-cache-disk-usage")
        cache_directory = SETTINGS.get("cache-directory")
        df = disk_usage(cache_directory)
        if df.percent > usage:
            LOG.debug("Cache disk usage %s, limit %s", df.percent, usage)
            self._housekeeping()
            delta = (df.percent - usage) * df.total * 0.01
            self._decache(delta)

    def _repr_html_(self):
        """Return a html representation of the cache .

        Returns
        -------
        str
            HTML status of the cache.
        """

        html = [css("table")]
        with self.new_connection() as db:
            for n in db.execute("SELECT * FROM cache"):
                n = dict(n)
                n["missing"] = not os.path.exists(n["path"])
                n["temporary"] = os.path.exists(n["path"] + ".tmp") or os.path.exists(
                    n["path"] + ".tmp.download"
                )  # TODO: decide how to handle temporary extension
                if n["size"] is None:
                    n["size"] = 0
                html.append("<table class='climetlab'>")
                html.append("<td><td colspan='2'>%s</td></tr>" % (n["path"],))

                for k in [x for x in n.keys() if x not in ("path", "owner_data")]:
                    v = humanize.bytes(n[k]) if k == "size" else n[k]
                    html.append("<td><td>%s</td><td>%s</td></tr>" % (k, v))
                html.append("</table>")
                html.append("<br>")
        return "".join(html)

    def _dump_cache_database(self, matcher=lambda x: True):
        result = []
        with self.connection as db:
            for d in db.execute("SELECT * FROM cache"):
                n = dict(d)
                for k in ("args", "owner_data"):
                    if n[k] is not None:
                        n[k] = json.loads(n[k])
                if matcher(n):
                    result.append(n)
        return result

    def _summary_dump_cache_database(self, matcher=lambda x: True):
        result = self._dump_cache_database(matcher=matcher)
        count = len(result)
        size = 0
        for r in result:
            size += r.get("size", 0)
        return count, size


CACHE = Cache()
CACHE.start()

dump_cache_database = in_executor(CACHE._dump_cache_database)
summary_dump_cache_database = in_executor(CACHE._summary_dump_cache_database)
register_cache_file = in_executor(CACHE._register_cache_file)
update_entry = in_executor(CACHE._update_entry)
check_cache_size = in_executor_forget(CACHE._check_cache_size)
cache_size = in_executor(CACHE._cache_size)
cache_entries = in_executor(CACHE._cache_entries)
purge_cache = in_executor(CACHE._purge_cache)
housekeeping = in_executor(CACHE._housekeeping)
decache_file = in_executor(CACHE._decache_file)
file_in_cache_directory = in_executor(CACHE._file_in_cache_directory)
settings_changed = in_executor(CACHE._settings_changed)
cache_directory = in_executor(CACHE._cache_directory)


def cache_file(
    owner: str,
    create,
    args,
    hash_extra=None,
    extension: str = ".cache",
    force=None,
    replace=None,
):
    """Creates a cache file in the climetlab cache-directory (defined in the :py:class:`Settings`).
    Uses :py:func:`_register_cache_file()`

    Parameters
    ----------
    owner : str
        The owner of the cache file is generally the name of the source that generated the cache.
    extension : str, optional
        Extension filename (such as ".nc" for NetCDF, etc.), by default ".cache"

    Returns
    -------
    path : str
        Full path to the cache file.
    """

    m = hashlib.sha256()
    m.update(owner.encode("utf-8"))

    m.update(
        json.dumps(args, sort_keys=True, default=default_serialiser).encode("utf-8")
    )
    m.update(json.dumps(hash_extra, sort_keys=True).encode("utf-8"))
    m.update(json.dumps(extension, sort_keys=True).encode("utf-8"))

    if replace is not None:
        # Don't replace files that are not in the cache
        if not file_in_cache_directory(replace):
            replace = None

    path = os.path.join(
        SETTINGS.get("cache-directory"),
        "{}-{}{}".format(
            owner.lower(),
            m.hexdigest(),
            extension,
        ),
    )

    try:
        record = register_cache_file(path, owner, args)
        if os.path.exists(path):
            if callable(force):
                owner_data = record["owner_data"]
                if owner_data is not None:
                    owner_data = json.loads(owner_data)
                force = force(args, path, owner_data)

            if force:
                decache_file(path)
    except Exception as e:
        LOG.warning("Cannot register cache file %s: %s", path, e)

    if not os.path.exists(path):
        lock = path + ".lock"

        with FileLock(lock):
            if not os.path.exists(
                path
            ):  # Check again, another thread/process may have created the file
                owner_data = create(path + ".tmp", args)

                os.rename(path + ".tmp", path)

                update_entry(path, owner_data)

                check_cache_size()

        try:
            os.unlink(lock)
        except OSError:
            pass

    return path


def auxiliary_cache_file(
    owner,
    path,
    index=0,
    content=None,
    extension=".cache",
):
    # Create an auxiliary cache file
    # to be used for example to cache an index
    # It is invalidated if `path` is changed
    stat = os.stat(path)

    def create(target, args):
        # Simply touch the file
        with open(target, "w") as f:
            if content:
                f.write(content)

    return cache_file(
        owner,
        create,
        (
            path,
            stat.st_ctime,
            stat.st_mtime,
            stat.st_size,
            index,
        ),
        extension=extension,
    )


# housekeeping()
SETTINGS.on_change(settings_changed)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import datetime
import itertools
import logging
import math
import os
import re
import time
from collections import defaultdict
from copy import deepcopy
from functools import cached_property

import numpy as np

from climetlab.core.order import build_remapping, normalize_order_by
from climetlab.utils import load_json_or_yaml
from climetlab.utils.humanize import seconds

LOG = logging.getLogger(__name__)


class DictObj(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for key, value in self.items():
            if isinstance(value, dict):
                self[key] = DictObj(value)
                continue
            if isinstance(value, list):
                self[key] = [
                    DictObj(item) if isinstance(item, dict) else item for item in value
                ]
                continue

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(attr)

    def __setattr__(self, attr, value):
        self[attr] = value


class Config(DictObj):
    def __init__(self, config):
        if isinstance(config, str):
            self.config_path = os.path.realpath(config)
            config = load_json_or_yaml(config)
        super().__init__(config)


class Inputs(list):
    def __init__(self, args):
        assert isinstance(args[0], (dict, Input)), args[0]
        args = [c if isinstance(c, Input) else Input(c) for c in args]
        super().__init__(args)

    def substitute(self, *args, **kwargs):
        return Inputs([i.substitute(*args, **kwargs) for i in self])

    def get_datetimes(self):
        # get datetime from each input
        # and make sure they are the same or None
        datetimes = None
        previous_name = None
        for i in self:
            new = i.get_datetimes()
            if new is None:
                continue
            new = sorted(list(new))
            if datetimes is None:
                datetimes = new

            if datetimes != new:
                raise ValueError(
                    "Mismatch in datetimes", previous_name, datetimes, i.name, new
                )
            previous_name = i.name

        if datetimes is None:
            raise ValueError(f"No datetimes found in {self}")

        return datetimes

    def do_load(self):
        datasets = {}
        for i in self:
            i = i.substitute(vars=datasets)
            ds = i.do_load()
            datasets[i.name] = ds

        out = None
        for ds in datasets.values():
            if out is None:
                out = ds
            else:
                out += ds

        from climetlab.readers.grib.index import FieldSet

        assert isinstance(out, FieldSet), type(out)

        return out

    def __repr__(self) -> str:
        return "\n".join(str(i) for i in self)


class DatetimeGetter:
    def __init__(self, kwargs) -> None:
        self.kwargs = kwargs

    @property
    def values(self):
        raise NotImplementedError(type(self))


class MarsDatetimeGetter(DatetimeGetter):
    def __init__(self, kwargs) -> None:
        super().__init__(kwargs)

        date = self.kwargs.get("date", [])
        hdate = self.kwargs.get("hdate", [])
        time = self.kwargs.get("time", [0])
        step = self.kwargs.get("step", [0])

        from climetlab.utils.dates import to_datetime_list

        date = to_datetime_list(date)
        hdate = to_datetime_list(hdate)
        time = make_list_int(time)
        step = make_list_int(step)

    pass


class StandardMarsDatetimeGetter(MarsDatetimeGetter):
    pass


class HindcastMarsDatetimeGetter(DatetimeGetter):
    pass


class Era5AccumulationDatetimeGetter(DatetimeGetter):
    pass


class ConstantDatetimeGetter(DatetimeGetter):
    @property
    def values(self):
        return None


class Input:
    _inheritance_done = False
    _inheritance_others = None
    _do_load = None

    def __init__(self, dic):
        assert isinstance(dic, dict), dic
        assert len(dic) == 1, dic

        self.name = list(dic.keys())[0]
        self.config = dic[self.name]

        if self.name == "forcing" or self.name == "constants":
            if "source_or_dataset" in self.config:
                # add $ to source_or_dataset for constants source.
                # climetlab will be refactored to remove this.
                assert self.config["source_or_dataset"][0] != "$", self.config[
                    "source_or_dataset"
                ]
                self.config["source_or_dataset"] = (
                    "$" + self.config["source_or_dataset"]
                )

        self.kwargs = self.config.get("kwargs", {})
        self.inherit = self.config.get("inherit", [])
        self.function = self.config.get("function", None)

    def get_datetimes(self):
        name = self.kwargs.get("name", None)

        assert name in [
            "era5-accumulations",
            "constants",
            "mars",
        ], f"{name} not implemented"

        if name == "constants":
            return None

        if name == "era5-accumulations":
            return None

        if name == "mars":
            is_hindast = "hdate" in self.kwargs

            date = self.kwargs.get("date", [])
            hdate = self.kwargs.get("hdate", [])
            time = self.kwargs.get("time", [0])
            step = self.kwargs.get("step", [0])

            from climetlab.utils.dates import to_datetime_list

            date = to_datetime_list(date)
            hdate = to_datetime_list(hdate)
            time = make_list_int(time)
            step = make_list_int(step)

            assert isinstance(date, (list, tuple)), date
            assert isinstance(time, (list, tuple)), time
            assert isinstance(step, (list, tuple)), step

            if is_hindast:
                assert isinstance(hdate, (list, tuple)), hdate
                if len(date) > 1 and len(hdate) > 1:
                    raise NotImplementedError(
                        (
                            f"Cannot have multiple dates in {self} "
                            "when using hindcast {date=}, {hdate=}"
                        )
                    )
                date = hdate
                del hdate

            if len(step) > 1 and len(time) > 1:
                raise NotImplementedError(
                    f"Cannot have multiple steps and multiple times in {self}"
                )

            datetimes = set()
            for d, t, s in itertools.product(date, time, step):
                new = build_datetime(date=d, time=t, step=s)
                if new in datetimes:
                    raise DuplicateDateTimeError(
                        f"Duplicate datetime '{new}' when processing << {self} >> already in {datetimes}"
                    )
                datetimes.add(new)
            return sorted(list(datetimes))

        raise ValueError(f"{name=} Cannot count number of elements in {self}")

    def do_load(self, others={}):
        if not self._do_load:
            from climetlab import load_dataset, load_source

            func = {
                None: load_source,
                "load_source": load_source,
                "load_dataset": load_dataset,
            }[self.function]

            ds = func(**self.kwargs)

            print(f"  Loading {self.name} of len {len(ds)}: {ds}")
            self._do_load = ds
        return self._do_load

    def get_first_field(self):
        return self.do_load()[0]

    def process_inheritance(self, others):
        for o in others:
            if o == self:
                continue
            name = o.name
            if name.startswith("$"):
                name = name[1:]
            if name not in self.inherit:
                continue
            if not o._inheritance_done:
                o.process_inheritance(others)

            kwargs = {}
            kwargs.update(o.kwargs)
            kwargs.update(self.kwargs)  # self.kwargs has priority
            self.kwargs = kwargs

        self._inheritance_others = others
        self._inheritance_done = True

    def __repr__(self) -> str:
        def repr(v):
            if isinstance(v, list):
                return f"{'/'.join(str(x) for x in v)}"
            return str(v)

        details = ", ".join(f"{k}={repr(v)}" for k, v in self.kwargs.items())
        return f"Input({self.name}, {details})<{self.inherit}"

    def substitute(self, *args, **kwargs):
        new_kwargs = substitute(self.kwargs.copy(), *args, **kwargs)
        i = Input(
            {
                self.name: dict(
                    kwargs=new_kwargs,
                    inherit=self.inherit,
                    function=self.function,
                )
            }
        )
        # if self._inheritance_others:
        #    i.process_inheritance(self._inheritance_others)
        return i


def make_list_int(value):
    if isinstance(value, str):
        if "/" not in value:
            return [value]
        bits = value.split("/")
        if len(bits) == 3 and bits[1].lower() == "to":
            value = list(range(int(bits[0]), int(bits[2]) + 1, 1))

        elif len(bits) == 5 and bits[1].lower() == "to" and bits[3].lower() == "by":
            value = list(range(int(bits[0]), int(bits[2]) + int(bits[4]), int(bits[4])))

    if isinstance(value, list):
        return value
    if isinstance(value, tuple):
        return value

    raise ValueError(f"Cannot make list from {value}")


def build_datetime(date, time, step):
    if isinstance(date, str):
        from climetlab.utils.dates import to_datetime

        date = to_datetime(date)

    if isinstance(time, int):
        if time < 24:
            time = f"{time:02d}00"
        else:
            time = f"{time:04d}"

    assert isinstance(date, datetime.datetime), date
    assert date.hour == 0 and date.minute == 0 and date.second == 0, date

    assert isinstance(time, str), time
    assert len(time) == 4, time
    assert int(time) >= 0 and int(time) < 2400, time
    if 0 < int(time) < 100:
        print(f"WARNING: {time=}, using time with minutes is unusual.")

    dt = datetime.datetime(
        year=date.year,
        month=date.month,
        day=date.day,
        hour=int(time[0:2]),
        minute=int(time[2:4]),
    )

    if step:
        dt += datetime.timedelta(hours=step)

    return dt


class InputHandler:
    def __init__(self, args, input, output):
        inputs = Inputs(input)
        self.output = output
        self.loops = [
            c
            if isinstance(c, Loop) and c.inputs == inputs
            else Loop(c, inputs, parent=self)
            for c in args
        ]
        if not self.loops:
            raise NotImplementedError("No loop")

    def iter_cubes(self):
        for loop in self.loops:
            yield from loop.iterate()

    @cached_property
    def n_iter_loops(self):
        return sum([loop.n_iter_loops for loop in self.loops])

    @property
    def first_cube(self):
        return self.first_cube_creator.to_cube()

    @property
    def first_cube_creator(self):
        for loop in self.loops:
            for cube_creator in loop.iterate():
                return cube_creator

    @property
    def chunking(self):
        return self.first_cube.chunking(self.output.chunking)

    @cached_property
    def n_cubes(self):
        n = 0
        for loop in self.loops:
            for i in loop.iterate():
                n += 1
        return n

    @cached_property
    def _info(self):
        infos = [loop._info for loop in self.loops]

        # check all are the same
        ref = infos[0]
        for c in infos:
            assert (np.array(ref.grid_points) == np.array(c.grid_points)).all(), (
                "grid_points mismatch",
                c.grid_points,
                ref.grid_points,
                type(ref.grid_points),
            )
            assert ref.resolution == c.resolution, (
                "resolution mismatch",
                c.resolution,
                ref.resolution,
            )
            assert ref.variables == c.variables, (
                "variables mismatch",
                c.variables,
                ref.variables,
            )

        coords = deepcopy(ref.coords)
        assert (
            "valid_datetime" in coords
        ), f"valid_datetime not found in coords {coords}"
        coords["valid_datetime"] = self.get_datetimes()

        for info in infos:
            for name, values in info.coords.items():
                if name == "valid_datetime":
                    continue
                assert values == ref.coords[name], (values, ref.coords[name])

        return Info(
            ref.first_field,
            ref.grid_points,
            ref.resolution,
            coords,
            ref.variables,
            ref.data_request,
        )

    @property
    def first_field(self):
        return self._info.first_field

    @property
    def grid_points(self):
        return self._info.grid_points

    @property
    def resolution(self):
        return self._info.resolution

    @property
    def data_request(self):
        return self._info.data_request

    @property
    def coords(self):
        return self._info.coords

    @property
    def variables(self):
        return self._info.variables

    @property
    def shape(self):
        shape = [len(c) for c in self.coords.values()]

        field_shape = list(self.first_field.shape)
        if self.output.flatten_grid:
            field_shape = [math.prod(field_shape)]

        return shape + field_shape

    def get_datetimes(self):
        # merge datetimes from all loops and check there are no duplicates
        datetimes = set()
        for i in self.loops:
            assert isinstance(i, Loop), i
            new = i.get_datetimes()
            for d in new:
                assert d not in datetimes, (d, datetimes)
                datetimes.add(d)
        datetimes = sorted(list(datetimes))

        def check(datetimes):
            if not datetimes:
                raise ValueError("No datetimes found.")
            if len(datetimes) == 1:
                raise ValueError("Only one datetime found.")

            delta = None
            for i in range(1, len(datetimes)):
                new = (datetimes[i] - datetimes[i - 1]).total_seconds() / 3600
                if not delta:
                    delta = new
                    continue
                if new != delta:
                    raise ValueError(
                        f"Datetimes are not regularly spaced: "
                        f"delta={new} hours  (date {i-1}={datetimes[i-1]}  date {i}={datetimes[i]}) "
                        f"Expecting {delta} hours  (date {0}={datetimes[0]}  date {1}={datetimes[1]}) "
                    )

        check(datetimes)

        return datetimes

    @property
    def frequency(self):
        datetimes = self.get_datetimes()
        freq = (datetimes[1] - datetimes[0]).total_seconds() / 3600
        assert round(freq) == freq, freq
        assert int(freq) == freq, freq
        return int(freq)

    def __repr__(self):
        return "InputHandler\n  " + "\n  ".join(str(i) for i in self.loops)


class Loop(dict):
    def __init__(self, dic, inputs, parent=None):
        assert isinstance(dic, dict), dic
        assert len(dic) == 1, dic
        super().__init__(dic)

        self.parent = parent
        self.name = list(dic.keys())[0]
        self.config = deepcopy(dic[self.name])

        assert "applies_to" in self.config, self.config
        applies_to = self.config.pop("applies_to")
        self.applies_to_inputs = Inputs(
            [input for input in inputs if input.name in applies_to]
        )
        for i in self.applies_to_inputs:
            i.process_inheritance(self.applies_to_inputs)

        self.values = {}
        for k, v in self.config.items():
            self.values[k] = self.expand(v)

    def expand(self, values):
        return expand(values)

    def __repr__(self) -> str:
        def repr_lengths(v):
            return f"{','.join([str(len(x)) for x in v])}"

        lenghts = [f"{k}({repr_lengths(v)})" for k, v in self.values.items()]
        return f"Loop({self.name}, {','.join(lenghts)}) {self.config}"

    @cached_property
    def n_iter_loops(self):
        return len(list(itertools.product(*self.values.values())))

    def iterate(self):
        for items in itertools.product(*self.values.values()):
            yield CubeCreator(
                inputs=self.applies_to_inputs,
                vars=dict(zip(self.values.keys(), items)),
                loop_config=self.config,
                output=self.parent.output,
            )

    @property
    def first(self):
        return CubeCreator(
            inputs=self.applies_to_inputs,
            vars={k: lst[0] for k, lst in self.values.items() if lst},
            loop_config=self.config,
            output=self.parent.output,
        )

    @cached_property
    def _info(self):
        first_info = self.first._info
        coords = deepcopy(first_info.coords)
        assert (
            "valid_datetime" in coords
        ), f"valid_datetime not found in coords {coords}"
        coords["valid_datetime"] = self.get_datetimes()
        return Info(
            first_field=first_info.first_field,
            grid_points=first_info.grid_points,
            resolution=first_info.resolution,
            coords=coords,
            variables=first_info.variables,
            data_request=first_info.data_request,
        )

    def get_datetimes(self):
        # merge datetimes from all cubecreators and check there are no duplicates
        datetimes = set()

        for i in self.iterate():
            assert isinstance(i, CubeCreator), i
            new = i.get_datetimes()

            duplicates = datetimes.intersection(set(new))
            if duplicates:
                raise DuplicateDateTimeError(
                    f"{len(duplicates)} duplicated datetimes "
                    f"'{sorted(list(duplicates))[0]},...' when processing << {self} >>"
                )

            datetimes = datetimes.union(set(new))
        return sorted(list(datetimes))


class DuplicateDateTimeError(ValueError):
    pass


class CubeCreator:
    def __init__(self, inputs, vars, loop_config, output):
        self._loop_config = loop_config
        self._vars = vars
        self._inputs = inputs
        self.output = output

        self.inputs = inputs.substitute(vars=vars, ignore_missing=True)

    @property
    def length(self):
        return 1

    def __repr__(self) -> str:
        out = f"CubeCreator ({self.length}):\n"
        out += f" loop_config: {self._loop_config}"
        out += f" vars: {self._vars}\n"
        out += " Inputs:\n"
        for _i, i in zip(self._inputs, self.inputs):
            out += f"- {_i}\n"
            out += f"  {i}\n"
        return out

    def do_load(self):
        return self.inputs.do_load()

    def get_datetimes(self):
        return self.inputs.get_datetimes()

    def to_cube(self):
        cube, _ = self._to_data_and_cube()
        return cube

    def _to_data_and_cube(self):
        data = self.do_load()

        start = time.time()
        print("Sorting dataset", self.output.order_by, self.output.remapping)
        cube = data.cube(
            self.output.order_by,
            remapping=self.output.remapping,
            flatten_values=self.output.flatten_grid,
        )
        cube = cube.squeeze()
        print(f"Sorting done in {seconds(time.time()-start)}.")

        def check(actual_dic, requested_dic):
            assert self.output.statistics in actual_dic

            for key in set(list(actual_dic.keys()) + list(requested_dic.keys())):
                actual = actual_dic[key]
                requested = requested_dic[key]

                actual = list(actual)

                if requested == "ascending":
                    assert actual == sorted(
                        actual
                    ), f"Requested= {requested} Actual= {actual}"
                    continue
                assert actual == requested, f"Requested= {requested} Actual= {actual}"

        check(actual_dic=cube.user_coords, requested_dic=self.output.order_by)

        return cube, data

    @property
    def _info(self):
        cube, data = self._to_data_and_cube()

        first_field = data[0]
        data_request = self._get_data_request(data)
        grid_points = first_field.grid_points()
        resolution = first_field.resolution
        coords = cube.user_coords
        variables = list(coords[list(coords.keys())[1]])

        return Info(
            first_field, grid_points, resolution, coords, variables, data_request
        )

    def _get_data_request(self, data):
        date = None
        params_levels = defaultdict(set)
        params_steps = defaultdict(set)

        for field in data:
            if not hasattr(field, "as_mars"):
                continue
            if date is None:
                date = field.valid_datetime()
            if field.valid_datetime() != date:
                continue

            as_mars = field.as_mars()
            step = as_mars.get("step")
            levtype = as_mars.get("levtype", "sfc")
            param = as_mars["param"]
            levelist = as_mars.get("levelist", None)
            area = field.mars_area
            grid = field.mars_grid

            if levelist is None:
                params_levels[levtype].add(param)
            else:
                params_levels[levtype].add((param, levelist))

            if step:
                params_steps[levtype].add((param, step))

        def sort(old_dic):
            new_dic = {}
            for k, v in old_dic.items():
                new_dic[k] = sorted(list(v))
            return new_dic

        params_steps = sort(params_steps)
        params_levels = sort(params_levels)

        out = dict(
            param_level=params_levels, param_step=params_steps, area=area, grid=grid
        )
        return out


def _format_list(x):
    if isinstance(x, (list, tuple)):
        if isinstance(x[0], datetime.datetime):
            is_regular = True
            delta = x[1] - x[0]
            for prev, current in zip(x[:-1], x[1:]):
                if current - prev != delta:
                    is_regular = False
                    break
            if is_regular:
                return f"{_format_list(x[0])}/to/{_format_list(x[-1])}/by/{delta.total_seconds()/3600}"

        return "/".join(_format_list(_) for _ in x)

    if isinstance(x, datetime.datetime):
        return x.strftime("%Y%m%d.%H%M")
    return str(x)


class Info:
    def __init__(
        self, first_field, grid_points, resolution, coords, variables, data_request
    ):
        assert len(set(variables)) == len(variables), (
            "Duplicate variables",
            variables,
        )

        assert grid_points[0].shape == grid_points[1].shape, (
            grid_points[0].shape,
            grid_points[1].shape,
            grid_points[0],
            grid_points[1],
        )

        assert len(grid_points) == 2, grid_points

        expected = math.prod(first_field.shape)
        assert len(grid_points[0]) == expected, (len(grid_points[0]), expected)

        self.first_field = first_field
        self.grid_points = grid_points
        self.resolution = resolution
        self.coords = coords
        self.variables = variables
        self.data_request = data_request

    def __repr__(self):
        shape = (
            f"{','.join([str(len(v)) for k, v in self.coords.items()])},"
            + f"{','.join([str(_) for _ in self.first_field.shape])}"
        )
        shape = shape.rjust(20)
        return (
            f"Info(first_field={self.first_field}, "
            f"resolution={self.resolution}, "
            f"variables={'/'.join(self.variables)})"
            f" coords={', '.join([k + ':' + _format_list(v) for k, v in self.coords.items()])}"
            f" {shape}"
        )


class Purpose:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return str(self.name)

    def __call__(self, config):
        pass

    @classmethod
    def dict_to_str(cls, x):
        if isinstance(x, str):
            return x
        return list(x.keys())[0]


class NonePurpose(Purpose):
    def __call__(self, config):
        config.output.flatten_grid = config.output.get("flatten_grid", False)
        config.output.ensemble_dimension = config.output.get(
            "ensemble_dimension", False
        )


class AifsPurpose(Purpose):
    def __call__(self, config):
        def check_dict_value_and_set(dic, key, value):
            if key in dic:
                if dic[key] != value:
                    raise ValueError(
                        f"Cannot use {key}={dic[key]} with {self} purpose. Must use {value}."
                    )
            dic[key] = value

        def ensure_element_in_list(lst, elt, index):
            if elt in lst:
                assert lst[index] == elt
                return lst

            _lst = [self.dict_to_str(d) for d in lst]
            if elt in _lst:
                assert _lst[index] == elt
                return lst

            return lst[:index] + [elt] + lst[index:]

        check_dict_value_and_set(config.output, "flatten_grid", True)
        check_dict_value_and_set(config.output, "ensemble_dimension", 2)

        assert isinstance(config.output.order_by, (list, tuple)), config.output.order_by
        config.output.order_by = ensure_element_in_list(
            config.output.order_by, "number", config.output.ensemble_dimension
        )

        order_by = config.output.order_by
        assert len(order_by) == 3, order_by
        assert self.dict_to_str(order_by[0]) == "valid_datetime", order_by
        assert self.dict_to_str(order_by[2]) == "number", order_by


PURPOSES = {None: NonePurpose, "aifs": AifsPurpose}


class LoadersConfig(Config):
    def __init__(self, config, *args, **kwargs):
        super().__init__(config, *args, **kwargs)

        if "description" not in self:
            raise ValueError("Must provide a description in the config.")

        purpose = PURPOSES[self.get("purpose")](self.get("purpose"))
        purpose(self)

        if not isinstance(self.input, (tuple, list)):
            print(f"WARNING: {self.input=} is not a list")
            self.input = [self.input]

        if "order" in self.output:
            raise ValueError(f"Do not use 'order'. Use order_by in {config}")
        if "order_by" in self.output:
            self.output.order_by = normalize_order_by(self.output.order_by)

        self.output.remapping = self.output.get("remapping", {})
        self.output.remapping = build_remapping(self.output.remapping)

        self.output.chunking = self.output.get("chunking", {})
        self.output.dtype = self.output.get("dtype", "float32")

        self.reading_chunks = self.get("reading_chunks")
        assert "flatten_values" not in self.output
        assert "flatten_grid" in self.output

        # The axis along which we append new data
        # TODO: assume grid points can be 2d as well
        self.output.append_axis = 0

        assert "statistics" in self.output
        statistics_axis_name = self.output.statistics
        statistics_axis = -1
        for i, k in enumerate(self.output.order_by):
            if k == statistics_axis_name:
                statistics_axis = i

        assert (
            statistics_axis >= 0
        ), f"{self.output.statistics} not in {list(self.output.order_by.keys())}"

        self.statistics_names = self.output.order_by[statistics_axis_name]

        # TODO: consider 2D grid points
        self.statistics_axis = statistics_axis

    def input_handler(self):
        return InputHandler(self.loops, self.input, output=self.output)


def substitute(x, vars=None, ignore_missing=False):
    """Recursively substitute environment variables and dict values in a nested list ot dict of string.
    substitution is performed using the environment var (if UPPERCASE) or the input dictionary.


    >>> substitute({'bar': '$bar'}, {'bar': '43'})
    {'bar': '43'}

    >>> substitute({'bar': '$BAR'}, {'BAR': '43'})
    Traceback (most recent call last):
        ...
    KeyError: 'BAR'

    >>> substitute({'bar': '$BAR'}, ignore_missing=True)
    {'bar': '$BAR'}

    >>> os.environ["BAR"] = "42"
    >>> substitute({'bar': '$BAR'})
    {'bar': '42'}

    >>> substitute('$bar', {'bar': '43'})
    '43'

    >>> substitute('$hdates_from_date($date, 2015, 2018)', {'date': '2023-05-12'})
    '2015-05-12/2016-05-12/2017-05-12/2018-05-12'

    """
    if vars is None:
        vars = {}
    if isinstance(x, (tuple, list)):
        return [substitute(y, vars, ignore_missing=ignore_missing) for y in x]

    if isinstance(x, dict):
        return {
            k: substitute(v, vars, ignore_missing=ignore_missing) for k, v in x.items()
        }

    if isinstance(x, str):
        if "$" not in x:
            return x

        lst = []

        for i, bit in enumerate(re.split(r"(\$(\w+)(\([^\)]*\))?)", x)):
            i %= 4
            if i in [2, 3]:
                continue
            if i == 1:
                try:
                    if "(" in bit:
                        # substitute by a function
                        FUNCTIONS = dict(hdates_from_date=hdates_from_date)

                        pattern = r"\$(\w+)\(([^)]*)\)"
                        match = re.match(pattern, bit)
                        assert match, bit

                        function_name = match.group(1)
                        params = [p.strip() for p in match.group(2).split(",")]
                        params = [
                            substitute(p, vars, ignore_missing=ignore_missing)
                            for p in params
                        ]

                        bit = FUNCTIONS[function_name](*params)

                    elif bit.upper() == bit:
                        # substitute by the var env if $UPPERCASE
                        bit = os.environ[bit[1:]]
                    else:
                        # substitute by the value in the 'vars' dict
                        bit = vars[bit[1:]]
                except KeyError as e:
                    if not ignore_missing:
                        raise e

            if bit != x:
                bit = substitute(bit, vars, ignore_missing=ignore_missing)

            lst.append(bit)

        lst = [_ for _ in lst if _ != ""]
        if len(lst) == 1:
            return lst[0]

        out = []
        for elt in lst:
            # if isinstance(elt, str):
            #    elt = [elt]
            assert isinstance(elt, (list, tuple)), elt
            out += elt
        return out

    return x


def hdates_from_date(date, start_year, end_year):
    """
    Returns a list of dates in the format '%Y%m%d' between start_year and end_year (inclusive),
    with the year of the input date.

    Args:
        date (str or datetime): The input date.
        start_year (int): The start year.
        end_year (int): The end year.

    Returns:
        List[str]: A list of dates in the format '%Y%m%d'.
    """
    if not str(start_year).isdigit():
        raise ValueError(f"start_year must be an int: {start_year}")
    if not str(end_year).isdigit():
        raise ValueError(f"end_year must be an int: {end_year}")
    start_year = int(start_year)
    end_year = int(end_year)

    from climetlab.utils.dates import to_datetime

    if isinstance(date, (list, tuple)):
        if len(date) != 1:
            raise NotImplementedError(f"{date} should have only one element.")
        date = date[0]

    date = to_datetime(date)
    assert not (date.hour or date.minute or date.second), date

    hdates = [date.replace(year=year) for year in range(start_year, end_year + 1)]
    return "/".join(d.strftime("%Y-%m-%d") for d in hdates)


class Expand(list):
    def __init__(self, config, **kwargs):
        self._config = config
        self.kwargs = kwargs
        self.groups = []
        self.parse_config()

    def parse_config(self):
        from climetlab.utils.dates import to_datetime

        self.start = self._config.get("start")
        if self.start is not None:
            self.start = to_datetime(self.start)
        self.end = self._config.get("end", self._config.get("stop"))
        if self.end is not None:
            self.end = to_datetime(self.end)
        self.step = self._config.get("step", self._config.get("frequency", 1))
        self.group_by = self._config.get("group_by")


class HindcastExpand(Expand):
    def __init__(self, config, **kwargs):
        super().__init__(config, **kwargs)
        self.groups = [["todo", "todo"]]


class ValuesExpand(Expand):
    def __init__(self, config, **kwargs):
        super().__init__(config, **kwargs)
        values = self._config["values"]
        values = [[v] if not isinstance(v, list) else v for v in values]
        for v in self._config["values"]:
            if not isinstance(v, (tuple, list)):
                v = [v]
            self.groups.append(v)


class StartStopExpand(Expand):
    def __init__(self, config, **kwargs):
        super().__init__(config, **kwargs)

        x = self.start
        all = []
        while x <= self.end:
            all.append(x)
            x += self.step

        result = [list(g) for _, g in itertools.groupby(all, key=self.grouper_key)]
        self.groups = [[format(x) for x in g] for g in result]

    def parse_config(self):
        if "stop" in self._config:
            raise ValueError(f"Use 'end' not 'stop' in loop. {self._config}")
        super().parse_config()

    def format(self, x):
        return x


class DateStartStopExpand(StartStopExpand):
    def grouper_key(self, x):
        return {
            1: lambda x: 0,  # only one group
            None: lambda x: x,  # one group per value
            "monthly": lambda dt: (dt.year, dt.month),
            "daily": lambda dt: (dt.year, dt.month, dt.day),
            "MMDD": lambda dt: (dt.month, dt.day),
        }[self.group_by](x)

    def parse_config(self):
        super().parse_config()
        assert isinstance(self.start, datetime.date), (type(self.start), self.start)
        assert isinstance(self.end, datetime.date), (type(self.end), self.end)
        self.step = datetime.timedelta(days=self.step)

    def format(self, x):
        return x.isoformat()


class IntStartStopExpand(StartStopExpand):
    def grouper_key(self, x):
        return {
            1: lambda x: 0,  # only one group
            None: lambda x: x,  # one group per value
        }[self.group_by](x)


def _expand_class(values):
    if isinstance(values, list):
        values = {"values": values}

    assert isinstance(values, dict), values

    if isinstance(values.get("values"), list):
        assert len(values) == 1, f"No other config keys implemented. {values}"
        return ValuesExpand

    if values.get("type") == "hindcast":
        return HindcastExpand

    if start := values.get("start"):
        if isinstance(start, datetime.datetime):
            return DateStartStopExpand
        if values.get("group_by") in ["monthly", "daily"]:
            return DateStartStopExpand
        return IntStartStopExpand

    raise ValueError(f"Cannot expand loop from {values}")


def expand(values, **kwargs):
    cls = _expand_class(values)
    return cls(values, **kwargs).groups



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab import load_source

from . import Dataset


class WeatherBench(Dataset):
    """
    This is an attempt to reproduce this research: https://arxiv.org/abs/2002.00469.
    See https://raspstephan.github.io/blog/weatherbench/
    There is a notebook available at:
    https://binder.pangeo.io/v2/gh/pangeo-data/WeatherBench/master?filepath=quickstart.ipynb
    """

    home_page = "https://raspstephan.github.io/blog/weatherbench/"
    citation = """@article{rasp2020weatherbench,
  title={WeatherBench: A benchmark dataset for data-driven weather forecasting},
  author={Rasp, Stephan and Dueben, Peter D and Scher, Sebastian and Weyn,
  Jonathan A and Mouatadid, Soukayna and Thuerey, Nils},
  journal={arXiv preprint arXiv:2002.00469},
  year={2020}
}
"""

    def __init__(self, parameter="geopotential_500", resolution=5.625):
        # self.check_parameter(
        #     "parameter",
        #     parameter,
        #     "10m_u_component_of_wind",
        #     "10m_v_component_of_wind",
        #     "2m_temperature",
        #     # "constants",
        #     "geopotential",
        #     "geopotential_500",
        #     "potential_vorticity",
        #     "relative_humidity",
        #     "specific_humidity",
        #     "temperature",
        #     "temperature_850",
        #     "toa_incident_solar_radiation",
        #     "total_cloud_cover",
        #     "total_precipitation",
        #     "u_component_of_wind",
        #     "v_component_of_wind",
        #     "vorticity",
        # )

        # self.check_parameter("resolution", resolution, 1.40625, 2.8125, 5.625)

        url = (
            "https://dataserv.ub.tum.de/s/m1524895"
            "/download?path=%2F{resolution}deg%2F{parameter}&files={parameter}_{resolution}deg.zip"
        ).format(resolution=resolution, parameter=parameter)
        self.source = load_source("url", url, verify=False)


dataset = WeatherBench



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import numpy as np

from .tensorflow import default_merger, to_funcs

LOG = logging.getLogger(__name__)


def to_pytorch(
    total_size=None,
    features=None,
    targets=None,
    options=None,
    targets_options=None,
    merger=default_merger,
    targets_merger=default_merger,
    #
    num_parallel_calls=10,
    prefetch=1024,
    shuffle_buffer_size=100,
    **kwargs,
):
    if total_size is None:
        total_size = len(features[0])
        print("totalsize", total_size)

    import torch

    func, func_targets = to_funcs(
        features, targets, options, targets_options, merger, targets_merger
    )

    class ClimetlabTorchDataset(torch.utils.data.Dataset):
        def __len__(self):
            """Returns the length of the dataset. Pytorch must know this."""
            return total_size

        def __getitem__(self, i):  # -> Tuple[np.ndarray, ...]:
            """Returns the i-th sample (x, y). Pytorch will take care of the shuffling after each epoch."""
            x = func(i)
            Y = func_targets(i)
            x = x.astype(np.float32)  # not needed ?
            Y = Y.astype(np.float32)  # not needed ?
            return x, Y

        # def __iter__(self):
        #    raise NotImplementedError('TODO')

    return ClimetlabTorchDataset()


class PytorchMixIn:
    def to_pytorch_dataloader(self, *args, dataloader_kwargs=None, **kwargs):
        from climetlab.ml.torch import to_pytorch_dataloader

        if dataloader_kwargs is None:
            dataloader_kwargs = {}
        dataset = self.to_pytorch(*args, **kwargs)
        return to_pytorch_dataloader(dataset, **dataloader_kwargs)

    def to_pytorch(self, *args, **kwargs):
        if "features" not in kwargs:
            kwargs["features"] = [self]

        return to_pytorch(*args, **kwargs)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import json
import logging
import os

from climetlab.core.caching import auxiliary_cache_file
from climetlab.readers.grib.codes import get_messages_positions
from climetlab.readers.grib.index import FieldSetInFiles
from climetlab.utils.parts import Part

LOG = logging.getLogger(__name__)


class FieldSetInOneFile(FieldSetInFiles):
    VERSION = 1

    @property
    def availability_path(self):
        return os.path.join(self.path, ".availability.pickle")

    def __init__(self, path, **kwargs):
        assert isinstance(path, str), path

        self.path = path
        self.offsets = None
        self.lengths = None
        self.mappings_cache_file = auxiliary_cache_file(
            "grib-index",
            path,
            content="null",
            extension=".json",
        )

        if not self._load_cache():
            self._build_offsets_lengths_mapping()

        super().__init__(**kwargs)

    def _build_offsets_lengths_mapping(self):
        offsets = []
        lengths = []

        for offset, length in get_messages_positions(self.path):
            offsets.append(offset)
            lengths.append(length)

        self.offsets = offsets
        self.lengths = lengths

        self._save_cache()

    def _save_cache(self):
        try:
            with open(self.mappings_cache_file, "w") as f:
                json.dump(
                    dict(
                        version=self.VERSION,
                        offsets=self.offsets,
                        lengths=self.lengths,
                    ),
                    f,
                )
        except Exception:
            LOG.exception("Write to cache failed %s", self.mappings_cache_file)

    def _load_cache(self):
        try:
            with open(self.mappings_cache_file) as f:
                c = json.load(f)
                if not isinstance(c, dict):
                    return False

                assert c["version"] == self.VERSION
                self.offsets = c["offsets"]
                self.lengths = c["lengths"]
                return True
        except Exception:
            LOG.exception("Load from cache failed %s", self.mappings_cache_file)

        return False

    def part(self, n):
        return Part(self.path, self.offsets[n], self.lengths[n])

    def number_of_parts(self):
        return len(self.offsets)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

# This module is called aaa so isort keeps it at the top on the imports
# as it needs to run first

import inspect
import os
import sys

LOADED_MODULES = {}
CLIMETLAB_DEBUG_IMPORTS = int(os.environ.get("CLIMETLAB_DEBUG_IMPORTS", 0))


class Requested:
    def find_spec(self, name, path=None, target=None):
        LOADED_MODULES[name] = []
        if CLIMETLAB_DEBUG_IMPORTS:
            for f in inspect.stack():
                if f.filename == __file__:
                    continue

                if "importlib._bootstrap" in f.filename:
                    continue

                LOADED_MODULES[name].append(f"{f.filename}:{f.lineno}")


class NotFound:
    def find_spec(self, name, path=None, target=None):
        LOADED_MODULES.pop(name, None)


def loaded_modules():
    # global LOADED_MODULES
    # return LOADED_MODULES
    return set(sys.modules.keys())


sys.meta_path.insert(0, Requested())
sys.meta_path.append(NotFound())



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import itertools
from collections import defaultdict

import climetlab as cml
from climetlab import Source
from climetlab.decorators import normalize

mapping = [
    [-1, 18, 6],
    [-1, 18, 7],
    [-1, 18, 8],
    [-1, 18, 9],
    [-1, 18, 10],
    [-1, 18, 11],
    [-1, 18, 12],
    [0, 6, 1],
    [0, 6, 2],
    [0, 6, 3],
    [0, 6, 4],
    [0, 6, 5],
    [0, 6, 6],
    [0, 6, 7],
    [0, 6, 8],
    [0, 6, 9],
    [0, 6, 10],
    [0, 6, 11],
    [0, 6, 12],
    [0, 18, 1],
    [0, 18, 2],
    [0, 18, 3],
    [0, 18, 4],
    [0, 18, 5],
]


class Era5Accumulations(Source):
    """This source use a MARS request to get the data for accumulated fields
    following the same logic as what has been done when storing ERA5 in the
    Climate Data Store (CDS).
    The date+time provided to this source refers to the valid datetime of the data,
    these date+time are used to compute the date+time+step that are actually used to perform the mars request.

    Note 1 : as all these are lists, there are potentially to many fields requested to MARS.
          This is taken care by the final ".sel" .
    Note 2 : There are no overlap due to the way the date+time+step are computed
    """

    def __init__(self, *args, **kwargs):
        request = {}
        for a in args:
            request.update(self.requests(**a))
        request = self.requests(**kwargs)

        user_dates = request["date"]
        user_times = request["time"]

        requested = set()

        dates = set()
        times = set()
        steps = set()

        for user_date, user_time in itertools.product(user_dates, user_times):
            assert isinstance(user_date, datetime.datetime), (
                type(user_date),
                user_dates,
                user_times,
            )
            assert isinstance(user_time, int), (type(user_time), user_dates, user_times)
            assert 0 <= user_time <= 24, user_time

            date = user_date + datetime.timedelta(hours=user_time)
            delta, time, step = mapping[date.hour]

            assert 0 <= time <= 23, time
            assert 0 <= step <= 24, step

            when = date + datetime.timedelta(days=delta)
            dates.add(datetime.datetime(when.year, when.month, when.day))
            times.add(time)
            steps.add(step)
            requested.add(date)

        valids = defaultdict(list)
        for date, time, step in itertools.product(dates, times, steps):
            valids[
                date + datetime.timedelta(hours=time) + datetime.timedelta(hours=step)
            ].append((date, time, step))

        got = set(valids.keys())
        assert all(len(x) == 1 for x in valids.values())
        missing = requested - got
        assert len(missing) == 0

        # extra = got - requested

        era_request = dict(**request)

        era_request.update(
            {
                "class": "ea",
                "type": "fc",
                "levtype": "sfc",
                "date": [d.strftime("%Y-%m-%d") for d in dates],
                "time": sorted(times),
                "step": sorted(steps),
            }
        )

        ds = cml.load_source("mars", **era_request)
        index = [d.valid_datetime() in requested for d in ds]
        self.ds = ds[index]

    def mutate(self):
        return self.ds

    @normalize("date", "date-list(datetime.datetime)")
    @normalize("time", "int-list")
    @normalize("area", "bounding-box(list)")
    def requests(self, **kwargs):
        result = dict(**kwargs)

        return result


source = Era5Accumulations



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

# The code is copied from skinnywms, and we should combile later

import datetime
from contextlib import closing
from itertools import product

import numpy as np

from climetlab.core import Base
from climetlab.utils.bbox import BoundingBox
from climetlab.utils.dates import to_datetime

from . import Reader


def as_datetime(self, time):
    return datetime.datetime.strptime(str(time)[:19], "%Y-%m-%dT%H:%M:%S")


def as_level(self, level):
    n = float(level)
    if int(n) == n:
        return int(n)
    return n


class Slice:
    def __init__(self, name, value, index, is_dimension, is_info):
        self.name = name
        self.index = index
        self.value = value
        self.is_dimension = (not is_info,)
        self.is_info = is_info

    def __repr__(self):
        return "[%s:%s=%s]" % (self.name, self.index, self.value)


class TimeSlice(Slice):
    pass


class Coordinate:
    def __init__(self, variable, info):
        self.variable = variable
        # We only support 1D coordinate for now
        # assert len(variable.dims) == 1
        self.is_info = info
        self.is_dimension = not info

        if variable.values.ndim == 0:
            self.values = [self.convert(variable.values)]
        else:
            self.values = [self.convert(t) for t in variable.values.flatten()]

    def make_slice(self, value):
        return self.slice_class(
            self.variable.name,
            value,
            self.values.index(value),
            self.is_dimension,
            self.is_info,
        )

    def __repr__(self):
        return "%s[name=%s,values=%s]" % (
            self.__class__.__name__,
            self.variable.name,
            len(self.values),
        )


class TimeCoordinate(Coordinate):
    slice_class = TimeSlice
    is_dimension = True
    convert = as_datetime


class LevelCoordinate(Coordinate):
    # This class is just in case we want to specialise
    # 'level', othewise, it is the same as OtherCoordinate
    slice_class = Slice
    is_dimension = False
    convert = as_level


class OtherCoordinate(Coordinate):
    slice_class = Slice
    is_dimension = False
    convert = as_level


class DataSet:
    def __init__(self, ds):
        self._ds = ds
        self._bbox = {}
        self._cache = {}

    @property
    def data_vars(self):
        return self._ds.data_vars

    def __getitem__(self, key):
        if key not in self._cache:
            self._cache[key] = self._ds[key]
        return self._cache[key]

    def bbox(self, variable):
        data_array = self[variable]
        dims = data_array.dims

        lat = dims[-2]
        lon = dims[-1]

        if (lat, lon) not in self._bbox:
            dims = data_array.dims

            latitude = data_array[lat]
            longitude = data_array[lon]

            self._bbox[(lat, lon)] = (
                np.amax(latitude.data),
                np.amin(longitude.data),
                np.amin(latitude.data),
                np.amax(longitude.data),
            )

        return self._bbox[(lat, lon)]


class NetCDFField(Base):
    def __init__(self, path, ds, variable, slices, non_dim_coords):
        data_array = ds[variable]

        self.north, self.west, self.south, self.east = ds.bbox(variable)

        self.path = path
        self.variable = variable
        self.slices = slices
        self.non_dim_coords = non_dim_coords

        self.name = self.variable

        self.title = getattr(
            data_array,
            "long_name",
            getattr(data_array, "standard_name", self.variable),
        )

        self.time = non_dim_coords.get("valid_time", non_dim_coords.get("time"))

        # print('====', non_dim_coords)

        for s in self.slices:
            if isinstance(s, TimeSlice):
                self.time = s.value

            if s.is_info:
                self.title += " (" + s.name + "=" + str(s.value) + ")"

    def to_numpy(self, *args, **kwargs):
        raise Exception(self.path, self.variable, self.slices)

    def plot_map(self, backend):
        dimensions = dict((s.name, s.index) for s in self.slices)

        backend.bounding_box(
            north=self.north, south=self.south, west=self.west, east=self.east
        )

        backend.plot_netcdf(self.path, self.variable, dimensions)

    def __repr__(self):
        return "NetCDFField[%r,%r]" % (self.variable, self.slices)

    def to_bounding_box(self):
        return BoundingBox(
            north=self.north, south=self.south, east=self.east, west=self.west
        )


class NetCDFReader(Reader):
    def __init__(self, source, path):
        super().__init__(source, path)
        self.fields = None

    def _scan(self):
        if self.fields is None:
            self.fields = self.get_fields()

    def __repr__(self):
        return "NetCDFReader(%s)" % (self.path,)

    def __iter__(self):
        self._scan()
        return iter(self.fields)

    def __len__(self):
        self._scan()
        return len(self.fields)

    def __getitem__(self, n):
        self._scan()
        return self.fields[n]

    def get_fields(self):
        import xarray as xr

        with closing(
            xr.open_mfdataset(self.path, combine="by_coords")
        ) as ds:  # or nested
            return self._get_fields(DataSet(ds))

    def _get_fields(self, ds):  # noqa C901
        # Select only geographical variables
        has_lat = False
        has_lon = False

        fields = []

        skip = set()

        for name in ds.data_vars:
            v = ds[name]
            skip.update(getattr(v, "coordinates", "").split(" "))
            skip.update(getattr(v, "bounds", "").split(" "))
            skip.update(getattr(v, "grid_mapping", "").split(" "))

        for name in ds.data_vars:
            if name in skip:
                continue

            v = ds[name]

            coordinates = []

            # self.log.info('Scanning file: %s var=%s coords=%s', self.path, name, v.coords)

            info = [value for value in v.coords if value not in v.dims]
            non_dim_coords = {}
            for coord in v.coords:
                if coord not in v.dims:
                    non_dim_coords[coord] = ds[coord].values
                    continue

                c = ds[coord]

                # self.log.info("COORD %s %s %s %s", coord, type(coord), hasattr(c, 'calendar'), c)

                standard_name = getattr(c, "standard_name", None)
                axis = getattr(c, "axis", None)
                long_name = getattr(c, "long_name", None)

                use = False

                if (
                    standard_name in ("longitude", "projection_x_coordinate")
                    or (long_name == "longitude")
                    or (axis == "X")
                ):
                    has_lon = True
                    use = True

                if (
                    standard_name in ("latitude", "projection_y_coordinate")
                    or (long_name == "latitude")
                    or (axis == "Y")
                ):
                    has_lat = True
                    use = True

                # Of course, not every one sets the standard_name
                if standard_name in ("time", "forecast_reference_time") or axis == "T":
                    coordinates.append(TimeCoordinate(c, coord in info))
                    use = True

                # TODO: Support other level types
                if standard_name in (
                    "air_pressure",
                    "model_level_number",
                    "altitude",
                ):  # or axis == 'Z':
                    coordinates.append(LevelCoordinate(c, coord in info))
                    use = True

                if axis in ("X", "Y"):
                    use = True

                if not use:
                    coordinates.append(OtherCoordinate(c, coord in info))

            if not (has_lat and has_lon):
                # self.log.info("NetCDFReader: skip %s (Not a 2 field)", name)
                continue

            for values in product(*[c.values for c in coordinates]):
                slices = []
                for value, coordinate in zip(values, coordinates):
                    slices.append(coordinate.make_slice(value))

                fields.append(NetCDFField(self.path, ds, name, slices, non_dim_coords))

        if not fields:
            raise Exception("NetCDFReader no 2D fields found in %s" % (self.path,))

        return fields

    def to_xarray(self, **kwargs):
        return type(self).to_xarray_multi_from_paths([self.path], **kwargs)

    @classmethod
    def to_xarray_multi_from_paths(cls, paths, **kwargs):
        import xarray as xr

        options = dict()
        options.update(kwargs.get("xarray_open_mfdataset_kwargs", {}))

        return xr.open_mfdataset(
            paths,
            **options,
        )

    def to_metview(self):
        from climetlab.metview import mv_read

        return mv_read(self.path)

    def plot_map(self, *args, **kwargs):
        return self.get_fields()[0].plot_map(*args, **kwargs)

    # Used by normalisers
    def to_datetime(self):
        times = self.to_datetime_list()
        assert len(times) == 1
        return times[0]

    def to_datetime_list(self):
        # TODO: check if that can be done faster
        result = set()
        for s in self.get_fields():
            result.add(to_datetime(s.time))
        return sorted(result)

    def to_bounding_box(self):
        return BoundingBox.multi_merge([s.to_bounding_box() for s in self.get_fields()])


def reader(source, path, magic=None, deeper_check=False):
    if magic is None or magic[:4] in (b"\x89HDF", b"CDF\x01", b"CDF\x02"):
        return NetCDFReader(source, path)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import math
import warnings

from climetlab.utils.kwargs import Kwargs
from climetlab.utils.serialise import deserialise_state, serialise_state

LOG = logging.getLogger(__name__)


class ItemWrapperForCfGrib:
    def __init__(self, item, ignore_keys=[]):
        self.item = item
        self.ignore_keys = ignore_keys

    def __getitem__(self, n):
        if n in self.ignore_keys:
            return None
        if n == "values":
            return self.item.values
        return self.item.metadata(n)


class IndexWrapperForCfGrib:
    def __init__(self, index=None, ignore_keys=[]):
        self.index = index
        self.ignore_keys = ignore_keys

    def __getstate__(self):
        return dict(index=serialise_state(self.index), ignore_keys=self.ignore_keys)

    def __setstate__(self, state):
        self.index = deserialise_state(state["index"])
        self.ignore_keys = state["ignore_keys"]

    def __getitem__(self, n):
        return ItemWrapperForCfGrib(
            self.index[n],
            ignore_keys=self.ignore_keys,
        )

    def __len__(self):
        return len(self.index)


class XarrayMixIn:
    def xarray_open_dataset_kwargs(self):
        return dict(
            cache=True,  # Set to false to prevent loading the whole dataset
            chunks=None,  # Set to 'auto' for lazy loading
        )

    def to_xarray(self, **kwargs):
        import xarray as xr

        xarray_open_dataset_kwargs = {}

        if "xarray_open_mfdataset_kwargs" in kwargs:
            warnings.warn(
                "xarray_open_mfdataset_kwargs is deprecated, please use xarray_open_dataset_kwargs instead."
            )
            kwargs["xarray_open_dataset_kwargs"] = kwargs.pop(
                "xarray_open_mfdataset_kwargs"
            )

        user_xarray_open_dataset_kwargs = kwargs.get("xarray_open_dataset_kwargs", {})

        # until ignore_keys is included into cfgrib,
        # it is implemented here directly
        ignore_keys = user_xarray_open_dataset_kwargs.get("backend_kwargs", {}).pop(
            "ignore_keys", []
        )

        for key in ["backend_kwargs"]:
            xarray_open_dataset_kwargs[key] = Kwargs(
                user=user_xarray_open_dataset_kwargs.pop(key, {}),
                default={"errors": "raise"},
                forced={},
                logging_owner="xarray_open_dataset_kwargs",
                logging_main_key=key,
            )

        default = dict(squeeze=False)  # TODO:Documenet me
        default.update(self.xarray_open_dataset_kwargs())

        xarray_open_dataset_kwargs.update(
            Kwargs(
                user=user_xarray_open_dataset_kwargs,
                default=default,
                forced={
                    "errors": "raise",
                    "engine": "cfgrib",
                },
            )
        )

        result = xr.open_dataset(
            IndexWrapperForCfGrib(self, ignore_keys=ignore_keys),
            **xarray_open_dataset_kwargs,
        )

        def math_prod(lst):
            if not hasattr(math, "prod"):
                # python 3.7 does not have math.prod
                n = 1
                for x in lst:
                    n = n * x
                return n
            return math.prod(lst)

        def number_of_gribs(da):
            # Assumes last two dimensions are lat/lon coordinates
            skip = 2
            if da.dims[-1] == "values":
                # Assumes last dimension is the one-dimensional
                # lat/lon coordinate (non-regular grid)
                skip = 1
            return math_prod(list(da.shape)[:-skip])

        two_d_fields = sum(number_of_gribs(result[v]) for v in result.data_vars)

        # Make sure all the fields are converted
        # There may be more 2D xarray fields than GRB fields
        # if some missing dimension are filled with NaN values

        assert two_d_fields >= len(self), (
            "Not all GRIB fields were converted to xarray"
            f" ({len(self)} GRIBs > {two_d_fields} 2D-field(s) in xarray)"
        )

        return result



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from . import Reader


class MatlabReader(Reader):
    def __init__(self, source, path):
        super().__init__(source, path)

    def to_numpy(self, key=None, scipy_io_loadmat_kwargs={}):
        from scipy.io import loadmat

        ds = loadmat(self.path, **scipy_io_loadmat_kwargs)
        if key is not None:
            ds = ds[key]
        return ds

    def __iter__(self):
        return iter([self])


def reader(source, path, magic=None, deeper_check=False):
    if magic[:7] == b"MATLAB ":
        return MatlabReader(source, path)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import glob
import logging
import os

from climetlab import load_source
from climetlab.core.settings import SETTINGS
from climetlab.readers import reader

from . import Source

LOG = logging.getLogger(__name__)


class FileSourceMeta(type(Source), type(os.PathLike)):
    def patch(cls, obj, *args, **kwargs):
        if "reader" in kwargs:
            setattr(obj, "reader", kwargs.pop("reader"))

        return super().patch(obj, *args, **kwargs)


class FileSource(Source, os.PathLike, metaclass=FileSourceMeta):
    _reader_ = None

    def __init__(self, path=None, filter=None, merger=None):
        self.path = path
        self.filter = filter
        self.merger = merger

    def mutate(self):
        if isinstance(self.path, (list, tuple)):
            if len(self.path) == 1:
                self.path = self.path[0]
            else:
                return load_source(
                    "multi",
                    [load_source("file", p) for p in self.path],
                    filter=self.filter,
                    merger=self.merger,
                )

        # Give a chance to directories and zip files
        # to return a multi-source
        source = self._reader.mutate_source()
        if source not in (None, self):
            source._parent = self
            return source
        return self

    def ignore(self):
        return self._reader.ignore()

    @classmethod
    def merge(cls, sources):
        from climetlab.mergers import merge_by_class

        assert all(isinstance(s, FileSource) for s in sources)
        return merge_by_class([s._reader for s in sources])

    @property
    def _reader(self):
        if self._reader_ is None:
            self._reader_ = reader(self, self.path)
        return self._reader_

    def __iter__(self):
        return iter(self._reader)

    def __len__(self):
        return len(self._reader)

    def __getitem__(self, n):
        return self._reader[n]

    def sel(self, *args, **kwargs):
        return self._reader.sel(*args, **kwargs)

    def order_by(self, *args, **kwargs):
        return self._reader.order_by(*args, **kwargs)

    def plot_map(self, *args, **kwargs):
        return self._reader.plot_map(*args, **kwargs)

    def to_xarray(self, **kwargs):
        return self._reader.to_xarray(**kwargs)

    def to_tfdataset(self, **kwargs):
        return self._reader.to_tfdataset(**kwargs)

    def to_pandas(self, **kwargs):
        LOG.debug("Calling reader.to_pandas %s", self)
        return self._reader.to_pandas(**kwargs)

    def to_numpy(self, **kwargs):
        return self._reader.to_numpy(**kwargs)

    def to_metview(self, **kwargs):
        return self._reader.to_metview(**kwargs)

    def save(self, path):
        return self._reader.save(path)

    def write(self, f):
        return self._reader.write(f)

    def scaled(self, *args, **kwargs):
        return self._reader.scaled(*args, **kwargs)

    def _attributes(self, names):
        return self._reader._attributes(names)

    def __repr__(self):
        cache_dir = SETTINGS.get("cache-directory")
        path = getattr(self, "path", None)
        if isinstance(path, str):
            path = path.replace(cache_dir, "CACHE:")
        try:
            reader_class_name = str(self._reader.__class__.__name__)
        except AttributeError as e:
            reader_class_name = str(e)
        except:  # noqa: E722
            reader_class_name = "Unknown"
        return f"{self.__class__.__name__}({path},{reader_class_name})"

    def __fspath__(self):
        return self.path

    # Used by normalisers
    def to_datetime(self):
        return self._reader.to_datetime()

    def to_datetime_list(self):
        return self._reader.to_datetime_list()

    def to_bounding_box(self):
        return self._reader.to_bounding_box()

    def statistics(self, **kwargs):
        return self._reader.statistics(**kwargs)


class File(FileSource):
    def __init__(
        self,
        path,
        expand_user=True,
        expand_vars=False,
        unix_glob=True,
        recursive_glob=True,
        filter=None,
        merger=None,
    ):
        if expand_user:
            path = os.path.expanduser(path)

        if expand_vars:
            path = os.path.expandvars(path)

        if unix_glob and set(path).intersection(set("[]?*")):
            matches = glob.glob(path, recursive=recursive_glob)
            if len(matches) == 1:
                path = matches[0]
            if len(matches) > 1:
                path = sorted(matches)

        super().__init__(path, filter, merger)


source = File



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import bz2

from . import Reader
from . import reader as find_reader


class BZReader(Reader):
    def __init__(self, source, path):
        super().__init__(source, path)

        def uncompress(target, _):
            with open(target, "wb") as g:
                with bz2.open(path, "rb") as f:
                    while True:
                        chunk = f.read(1024 * 1204)
                        if not chunk:
                            break
                        g.write(chunk)

        self.unzipped = self.cache_file(
            uncompress,
            dict(path=path),
        )

    def mutate(self):
        print("mutare", self.source, self.unzipped)
        return find_reader(self.source, self.unzipped)


def reader(source, path, magic=None, deeper_check=False):
    if magic is None:  # Bypass check and force
        return BZReader(source, path)

    if magic[:3] == b"BZh":
        return BZReader(source, path)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.wrappers import get_wrapper


def _normalize(lon, minimum):
    while lon < minimum:
        lon += 360

    while lon >= minimum + 360:
        lon -= 360

    return lon


class BoundingBox:
    def __init__(self, *, north, west, south, east):
        # Convert to float as these values may come from Numpy
        self.north = min(float(north), 90.0)
        self.south = max(float(south), -90.0)

        self.is_periodic_west_east = (east - west) == 360
        self.west = _normalize(float(west), -180)  # Or 0?

        if self.is_periodic_west_east:
            self.east = self.west + 360
        else:
            self.east = _normalize(float(east), self.west)

        if self.north < self.south:
            raise ValueError(
                f"Invalid bounding box, north={self.north} < south={self.south}"
            )

        if self.west > self.east:
            raise ValueError(
                f"Invalid bounding box, west={self.west} > east={self.east}"
            )

        if self.east > self.west + 360:
            raise ValueError(
                f"Invalid bounding box, east={self.east} > west={self.west}+360"
            )

    def __repr__(self):
        return "BoundingBox(north=%g,west=%g,south=%g,east=%g)" % (
            self.north,
            self.west,
            self.south,
            self.east,
        )

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False

        return self.as_tuple() == other.as_tuple()

    @property
    def width(self):
        return self.east - self.west

    @property
    def height(self):
        return self.north - self.south

    @classmethod
    def multi_merge(cls, bboxes):
        north = max(z.north for z in bboxes)
        south = min(z.south for z in bboxes)

        first = bboxes[0]

        origin = first.east % 360
        full = BoundingBox(
            north=north,
            west=first.west,
            south=south,
            east=first.west + 360,
        )

        # Build the list of provided [west, east] intervals

        boundaries = list()
        stacked_intervals = set()  # To keep track of overlapping intervals
        for i, b in enumerate(bboxes):
            if b.east - b.west == 360:
                return full

            west = (b.west - origin) % 360
            east = (b.east - origin) % 360

            if west > east:
                stacked_intervals.add(i)

            boundaries.append((west, True, i))
            boundaries.append((east, False, i))

        boundaries = sorted(boundaries)

        start = 0
        best = -1
        west = 0
        east = 0

        # Find the longest interval *outside* the provided [west, east] intervals
        # It's complement will be the result

        for cursor, entering, interval in boundaries:
            if entering:
                if not stacked_intervals:
                    distance = cursor - start
                    if distance > best:
                        best = distance
                        west = cursor
                        east = start

                    start = None

                stacked_intervals.add(interval)

            else:  # exiting
                stacked_intervals.remove(interval)
                if not stacked_intervals:
                    start = cursor

        if best <= 0:
            return full

        return BoundingBox(
            north=north,
            west=origin + west,
            south=south,
            east=origin + east,
        )

    def merge(self, other):
        return self.multi_merge([self, other])

    def add_margins(self, margins):
        if isinstance(margins, str) and margins[-1] == "%":
            margins = int(margins[:-1]) / 100.0
            margins = max(
                (self.north - self.south) * margins, (self.east - self.west) * margins
            )

        # TODO:check east/west
        margins_lat = margins
        margins_lon = margins

        if self.east - self.west > 360 - 2 * margins:
            margins = (360 - (self.east - self.west)) / 2.0

        return BoundingBox(
            north=self.north + margins_lat,
            west=self.west - margins_lon,
            south=self.south - margins_lat,
            east=self.east + margins_lon,
        )

    def as_list(self):
        return [self.north, self.west, self.south, self.east]

    def as_tuple(self):
        return (self.north, self.west, self.south, self.east)

    def as_dict(self):
        return dict(north=self.north, west=self.west, south=self.south, east=self.east)


def to_bounding_box(obj):
    if isinstance(obj, BoundingBox):
        return obj

    if isinstance(obj, (list, tuple)):
        return BoundingBox(north=obj[0], west=obj[1], south=obj[2], east=obj[3])

    obj = get_wrapper(obj)

    return to_bounding_box(obj.to_bounding_box())
